"""Tests for complex commands executor."""

from typing import Literal, Union, cast

import pytest
from decoy import Decoy, matchers

from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    BlowoutLocation,
    Coordinate,
    LiquidClassSchemaV1,
    PositionReference,
)
from opentrons_shared_data.liquid_classes.types import TipPositionDict

from opentrons.protocol_api import Labware, TrashBin, WasteChute
from opentrons.protocol_api._liquid import LiquidClass
from opentrons.protocol_api._liquid_properties import TransferProperties
from opentrons.protocol_api.core.engine.instrument import InstrumentCore
from opentrons.protocol_api.core.engine.transfer_components_executor import (
    AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP,
    LiquidAndAirGapPair,
    TipState,
    TransferComponentsExecutor,
    TransferType,
    absolute_point_from_position_reference_and_offset,
)
from opentrons.protocol_api.core.engine.well import WellCore
from opentrons.protocol_api.disposal_locations import DisposalOffset
from opentrons.protocol_api.labware import Well
from opentrons.protocols.advanced_control.transfers import (
    transfer_liquid_utils as tx_utils,
)
from opentrons.protocols.advanced_control.transfers.transfer_liquid_utils import (
    LocationCheckDescriptors,
)
from opentrons.types import Location, Mount, Point


@pytest.fixture
def mock_instrument_core(decoy: Decoy) -> InstrumentCore:
    """Return a mocked out instrument core."""
    return decoy.mock(cls=InstrumentCore)


@pytest.fixture
def sample_transfer_props(
    maximal_liquid_class_def: LiquidClassSchemaV1,
) -> TransferProperties:
    """Return a mocked out liquid class fixture."""
    return LiquidClass.create(maximal_liquid_class_def).get_for(
        pipette="flex_1channel_50", tip_rack="opentrons_flex_96_tiprack_50ul"
    )


@pytest.fixture(autouse=True)
def patch_mock_raise_if_location_inside_liquid(
    decoy: Decoy, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Replace raise_if_location_inside_liquid() with a mock."""
    mock = decoy.mock(func=tx_utils.raise_if_location_inside_liquid)
    monkeypatch.setattr(tx_utils, "raise_if_location_inside_liquid", mock)


@pytest.fixture(autouse=True)
def patch_mock_get_blowout_location_for_trash(
    decoy: Decoy, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Replace get_blowout_location_for_trash() with a mock."""
    mock = decoy.mock(func=tx_utils.get_blowout_location_for_trash)
    monkeypatch.setattr(tx_utils, "get_blowout_location_for_trash", mock)


""" Test aspirate properties:
"submerge": {
  "startPosition": {"positionReference": "well-top", "offset": {"x": 1, "y": 2, "z": 3}},
  "speed": 100,
  "delay": {"enable": true, "params": {"duration": 10.0}}},
"retract": {
  "endPosition": {"positionReference": "well-top", "offset": {"x": 3, "y": 2, "z": 1}},
  "speed": 50,
  "airGapByVolume": [[1.0, 0.1], [49.9, 0.1], [50.0, 0.0]],
  "touchTip": {"enable": true, "params": {"zOffset": -1, "mmFromEdge": 0.5, "speed": 30}},
  "delay": {"enable": true, "params": {"duration": 20}}},
"aspiratePosition": {"positionReference": "well-bottom", "offset": {"x": 10, "y": 20, "z": 30}},
"flowRateByVolume": [[1.0, 35.0], [10.0, 24.0], [50.0, 35.0]],
"correctionByVolume": [[0.0, 0.0]],
"preWet": true,
"mix": {"enable": true, "params": {"repetitions": 1, "volume": 50}},
"delay": {"enable": true, "params": {"duration": 0.2}}
"""


@pytest.mark.parametrize(
    argnames=[
        "air_gap_volume",
        "air_gap_flow_rate_by_vol",
    ],
    argvalues=[(0.123, 123), (1.23, 0.123)],
)
def test_submerge(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
    air_gap_volume: float,
    air_gap_flow_rate_by_vol: float,
) -> None:
    """Should perform the expected submerge steps."""
    source_well = decoy.mock(cls=WellCore)
    well_top_point = Point(1, 2, 4)
    well_bottom_point = Point(4, 5, 6)
    air_gap_correction_by_vol = 0.321
    sample_transfer_props.dispense.flow_rate_by_volume.set_for_volume(
        air_gap_volume, air_gap_flow_rate_by_vol
    )
    sample_transfer_props.dispense.correction_by_volume.set_for_volume(
        air_gap_volume, air_gap_correction_by_vol
    )

    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=Location(Point(1, 2, 3), labware=None),
        target_well=source_well,
        tip_state=TipState(
            ready_to_aspirate=True,
            last_liquid_and_air_gap_in_tip=LiquidAndAirGapPair(
                liquid=0, air_gap=air_gap_volume
            ),
        ),
        transfer_type=TransferType.ONE_TO_ONE,
    )
    decoy.when(source_well.get_bottom(0)).then_return(well_bottom_point)
    decoy.when(source_well.get_top(0)).then_return(well_top_point)
    decoy.when(source_well.get_top(2)).then_return(Point(1, 2, 6))
    subject.submerge(
        submerge_properties=sample_transfer_props.aspirate.submerge,
        post_submerge_action="aspirate",
    )
    decoy.verify(
        tx_utils.raise_if_location_inside_liquid(
            location=Location(Point(x=2, y=4, z=7), labware=None),
            well_core=source_well,
            location_check_descriptors=LocationCheckDescriptors(
                location_type="submerge start",
                pipetting_action="aspirate",
            ),
            logger=matchers.Anything(),
        ),
        mock_instrument_core.move_to(
            location=Location(Point(x=2, y=4, z=7), labware=None),
            well_core=source_well,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
        ),
        mock_instrument_core.remove_air_gap_during_transfer_with_liquid_class(
            last_air_gap=air_gap_volume,
            dispense_props=sample_transfer_props.dispense,
            location=Location(Point(x=2, y=4, z=7), labware=None),
        ),
        mock_instrument_core.move_to(
            location=Location(Point(1, 2, 3), labware=None),
            well_core=source_well,
            force_direct=True,
            minimum_z_height=None,
            speed=100,
        ),
        mock_instrument_core.delay(10),
    )


def test_submerge_without_starting_air_gap(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
) -> None:
    """Should perform the expected submerge steps."""
    source_well = decoy.mock(cls=WellCore)
    well_top_point = Point(1, 2, 4)
    well_bottom_point = Point(4, 5, 6)
    air_gap_volume = 0
    sample_transfer_props.dispense.flow_rate_by_volume.set_for_volume(
        air_gap_volume, 1234
    )
    sample_transfer_props.dispense.correction_by_volume.set_for_volume(
        air_gap_volume, 1234
    )

    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=Location(Point(1, 2, 3), labware=None),
        target_well=source_well,
        tip_state=TipState(
            ready_to_aspirate=True,
            last_liquid_and_air_gap_in_tip=LiquidAndAirGapPair(
                liquid=0, air_gap=air_gap_volume
            ),
        ),
        transfer_type=TransferType.ONE_TO_ONE,
    )
    decoy.when(source_well.get_bottom(0)).then_return(well_bottom_point)
    decoy.when(source_well.get_top(0)).then_return(well_top_point)
    decoy.when(source_well.get_top(2)).then_return(Point(1, 2, 6))
    subject.submerge(
        submerge_properties=sample_transfer_props.aspirate.submerge,
        post_submerge_action="aspirate",
    )
    decoy.verify(
        tx_utils.raise_if_location_inside_liquid(
            location=Location(Point(x=2, y=4, z=7), labware=None),
            well_core=source_well,
            location_check_descriptors=LocationCheckDescriptors(
                location_type="submerge start",
                pipetting_action="aspirate",
            ),
            logger=matchers.Anything(),
        ),
        mock_instrument_core.move_to(
            location=Location(Point(x=2, y=4, z=7), labware=None),
            well_core=source_well,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
        ),
        mock_instrument_core.move_to(
            location=Location(Point(1, 2, 3), labware=None),
            well_core=source_well,
            force_direct=True,
            minimum_z_height=None,
            speed=100,
        ),
        mock_instrument_core.delay(10),
    )


@pytest.mark.parametrize(
    argnames=[
        "air_gap_volume",
        "air_gap_flow_rate_by_vol",
    ],
    argvalues=[(0.123, 123), (1.23, 0.123)],
)
def test_submerge_with_trash_location(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
    air_gap_volume: float,
    air_gap_flow_rate_by_vol: float,
) -> None:
    """Should perform the expected submerge steps."""
    air_gap_correction_by_vol = 0.321
    sample_transfer_props.dispense.flow_rate_by_volume.set_for_volume(
        air_gap_volume, air_gap_flow_rate_by_vol
    )
    sample_transfer_props.dispense.correction_by_volume.set_for_volume(
        air_gap_volume, air_gap_correction_by_vol
    )
    mock_trash_bin = decoy.mock(cls=TrashBin)

    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=mock_trash_bin,
        target_well=None,
        tip_state=TipState(
            ready_to_aspirate=True,
            last_liquid_and_air_gap_in_tip=LiquidAndAirGapPair(
                liquid=0, air_gap=air_gap_volume
            ),
        ),
        transfer_type=TransferType.ONE_TO_ONE,
    )
    subject.submerge(
        submerge_properties=sample_transfer_props.dispense.submerge,
        post_submerge_action="dispense",
    )

    decoy.verify(
        mock_instrument_core.move_to(
            location=mock_trash_bin,
            well_core=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
        ),
        mock_instrument_core.remove_air_gap_during_transfer_with_liquid_class(
            last_air_gap=air_gap_volume,
            dispense_props=sample_transfer_props.dispense,
            location=mock_trash_bin,
        ),
        mock_instrument_core.delay(1.1),
    )


def test_submerge_raises_when_submerge_point_is_invalid(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
) -> None:
    """Should raise an error when submerge start point is invalid."""
    source_well = decoy.mock(cls=WellCore)
    well_top_point = Point(1, 2, 4)
    well_bottom_point = Point(4, 5, 6)
    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=Location(Point(1, 2, 3), labware=None),
        target_well=source_well,
        tip_state=TipState(
            ready_to_aspirate=True,
            last_liquid_and_air_gap_in_tip=LiquidAndAirGapPair(liquid=0, air_gap=123),
        ),
        transfer_type=TransferType.ONE_TO_ONE,
    )
    decoy.when(source_well.get_bottom(0)).then_return(well_bottom_point)
    decoy.when(source_well.get_top(0)).then_return(well_top_point)
    decoy.when(mock_instrument_core.get_liquid_presence_detection()).then_return(False)
    decoy.when(
        tx_utils.raise_if_location_inside_liquid(
            location=Location(Point(x=2, y=4, z=7), labware=None),
            well_core=source_well,
            location_check_descriptors=LocationCheckDescriptors(
                location_type="submerge start",
                pipetting_action="aspirate",
            ),
            logger=matchers.Anything(),
        )
    ).then_raise(RuntimeError("Oh no!"))
    with pytest.raises(RuntimeError, match="Oh no!"):
        subject.submerge(
            submerge_properties=sample_transfer_props.aspirate.submerge,
            post_submerge_action="aspirate",
        )


@pytest.mark.parametrize(
    argnames=["position_reference"],
    argvalues=[
        [PositionReference.WELL_BOTTOM],
        [PositionReference.LIQUID_MENISCUS],
    ],
)
def test_aspirate_and_wait(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
    position_reference: PositionReference,
) -> None:
    """It should execute an aspirate and a delay according to properties."""
    source_well = decoy.mock(cls=WellCore)
    sample_transfer_props.aspirate.aspirate_position.position_reference = (
        position_reference
    )
    aspirate_flow_rate = (
        sample_transfer_props.aspirate.flow_rate_by_volume.get_for_volume(10)
    )
    decoy.when(mock_instrument_core.get_current_volume()).then_return(5)
    correction_volume = (
        sample_transfer_props.aspirate.correction_by_volume.get_for_volume(15)
    )
    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=Location(Point(1, 2, 3), labware=None),
        target_well=source_well,
        tip_state=TipState(),
        transfer_type=TransferType.ONE_TO_ONE,
    )
    subject.aspirate_and_wait(volume=10)
    decoy.verify(
        mock_instrument_core.aspirate(
            location=Location(Point(1, 2, 3), labware=None),
            well_core=None,
            volume=10,
            rate=1,
            flow_rate=aspirate_flow_rate,
            in_place=True,
            correction_volume=correction_volume,
        ),
        mock_instrument_core.delay(0.2),
    )


def test_aspirate_and_wait_skips_delay(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
) -> None:
    """It should skip the wait after aspirate."""
    sample_transfer_props.aspirate.delay.enabled = False
    source_well = decoy.mock(cls=WellCore)
    decoy.when(mock_instrument_core.get_current_volume()).then_return(5)

    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=Location(Point(1, 2, 3), labware=None),
        target_well=source_well,
        tip_state=TipState(),
        transfer_type=TransferType.ONE_TO_ONE,
    )
    subject.aspirate_and_wait(volume=10)
    decoy.verify(
        mock_instrument_core.delay(0.2),
        times=0,
    )


@pytest.mark.parametrize(
    argnames=["position_reference"],
    argvalues=[
        [PositionReference.WELL_BOTTOM],
        [PositionReference.LIQUID_MENISCUS],
    ],
)
def test_dispense_and_wait(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
    position_reference: PositionReference,
) -> None:
    """It should execute a dispense and a delay according to properties."""
    source_well = decoy.mock(cls=WellCore)
    sample_transfer_props.dispense.dispense_position.position_reference = (
        position_reference
    )
    dispense_flow_rate = (
        sample_transfer_props.dispense.flow_rate_by_volume.get_for_volume(10)
    )
    decoy.when(mock_instrument_core.get_current_volume()).then_return(50)
    correction_volume = (
        sample_transfer_props.dispense.correction_by_volume.get_for_volume(40)
    )
    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=Location(Point(1, 2, 3), labware=None),
        target_well=source_well,
        tip_state=TipState(),
        transfer_type=TransferType.ONE_TO_ONE,
    )
    subject.dispense_and_wait(
        dispense_properties=sample_transfer_props.dispense,
        volume=10,
        push_out_override=123,
    )
    decoy.verify(
        mock_instrument_core.dispense(
            location=Location(Point(1, 2, 3), labware=None),
            well_core=None,
            volume=10,
            rate=1,
            flow_rate=dispense_flow_rate,
            in_place=True,
            push_out=123,
            correction_volume=correction_volume,
        ),
        mock_instrument_core.delay(0.5),
    )


def test_dispense_and_wait_skips_delay(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
) -> None:
    """It should skip the wait after dispense."""
    sample_transfer_props.dispense.delay.enabled = False
    source_well = decoy.mock(cls=WellCore)
    decoy.when(mock_instrument_core.get_current_volume()).then_return(50)

    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=Location(Point(1, 2, 3), labware=None),
        target_well=source_well,
        tip_state=TipState(),
        transfer_type=TransferType.ONE_TO_ONE,
    )
    subject.dispense_and_wait(
        dispense_properties=sample_transfer_props.dispense,
        volume=10,
        push_out_override=123,
    )
    decoy.verify(
        mock_instrument_core.delay(0.2),
        times=0,
    )


def test_dispense_and_wait_raises_if_tip_volume_less_than_dispense_vol(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
) -> None:
    """Should raise a useful error if trying to dispense more than liquid present in tip."""
    decoy.when(mock_instrument_core.get_current_volume()).then_return(50)

    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=Location(Point(1, 2, 3), labware=None),
        target_well=decoy.mock(cls=WellCore),
        tip_state=TipState(),
        transfer_type=TransferType.ONE_TO_ONE,
    )
    with pytest.raises(
        RuntimeError, match="Cannot dispense 51uL when the tip has only 50uL."
    ):
        subject.dispense_and_wait(
            dispense_properties=sample_transfer_props.dispense,
            volume=51,
            push_out_override=123,
        )


def test_dispense_into_trash_and_wait(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
) -> None:
    """It should execute a dispense and a delay according to properties."""
    mock_trash_bin = decoy.mock(cls=TrashBin)
    dispense_flow_rate = (
        sample_transfer_props.dispense.flow_rate_by_volume.get_for_volume(10)
    )
    decoy.when(mock_instrument_core.get_current_volume()).then_return(50)
    correction_volume = (
        sample_transfer_props.dispense.correction_by_volume.get_for_volume(40)
    )
    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=mock_trash_bin,
        target_well=None,
        tip_state=TipState(),
        transfer_type=TransferType.ONE_TO_ONE,
    )
    subject.dispense_and_wait(
        dispense_properties=sample_transfer_props.dispense,
        volume=10,
        push_out_override=123,
    )
    decoy.verify(
        mock_instrument_core.dispense(
            location=mock_trash_bin,
            well_core=None,
            volume=10,
            rate=1,
            flow_rate=dispense_flow_rate,
            in_place=True,
            push_out=123,
            correction_volume=correction_volume,
        ),
        mock_instrument_core.delay(0.5),
    )


def test_mix(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
) -> None:
    """It should execute mix steps."""
    source_well = decoy.mock(cls=WellCore)
    aspirate_flow_rate = (
        sample_transfer_props.aspirate.flow_rate_by_volume.get_for_volume(50)
    )
    dispense_flow_rate = (
        sample_transfer_props.dispense.flow_rate_by_volume.get_for_volume(50)
    )
    decoy.when(mock_instrument_core.get_current_volume()).then_return(0, 50)
    aspirate_correction_volume = (
        sample_transfer_props.aspirate.correction_by_volume.get_for_volume(50)
    )
    dispense_correction_volume = (
        sample_transfer_props.dispense.correction_by_volume.get_for_volume(0)
    )
    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=Location(Point(1, 2, 3), labware=None),
        target_well=source_well,
        tip_state=TipState(),
        transfer_type=TransferType.ONE_TO_ONE,
    )
    subject.mix(
        mix_properties=sample_transfer_props.aspirate.mix,
        last_dispense_push_out=True,
    )

    decoy.verify(
        mock_instrument_core.aspirate(
            location=Location(Point(1, 2, 3), labware=None),
            well_core=None,
            volume=50,
            rate=1,
            flow_rate=aspirate_flow_rate,
            in_place=True,
            correction_volume=aspirate_correction_volume,
        ),
        mock_instrument_core.delay(0.2),
        mock_instrument_core.dispense(
            location=Location(Point(1, 2, 3), labware=None),
            well_core=None,
            volume=50,
            rate=1,
            flow_rate=dispense_flow_rate,
            in_place=True,
            push_out=2.0,
            correction_volume=dispense_correction_volume,
        ),
        mock_instrument_core.delay(0.5),
    )


def test_mix_disabled(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
) -> None:
    """It should not perform a mix when it is disabled."""
    sample_transfer_props.aspirate.mix.enabled = False
    source_well = decoy.mock(cls=WellCore)
    aspirate_flow_rate = (
        sample_transfer_props.aspirate.flow_rate_by_volume.get_for_volume(50)
    )
    correction_volume = (
        sample_transfer_props.aspirate.correction_by_volume.get_for_volume(50)
    )
    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=Location(Point(1, 2, 3), labware=None),
        target_well=source_well,
        tip_state=TipState(),
        transfer_type=TransferType.ONE_TO_ONE,
    )
    subject.mix(
        mix_properties=sample_transfer_props.aspirate.mix,
        last_dispense_push_out=True,
    )
    decoy.verify(
        mock_instrument_core.aspirate(
            location=Location(Point(1, 2, 3), labware=None),
            well_core=None,
            volume=50,
            rate=1,
            flow_rate=aspirate_flow_rate,
            in_place=True,
            correction_volume=correction_volume,
        ),
        times=0,
    )


def test_pre_wet(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
) -> None:
    """It should execute pre-wet steps."""
    source_well = decoy.mock(cls=WellCore)
    aspirate_flow_rate = (
        sample_transfer_props.aspirate.flow_rate_by_volume.get_for_volume(40)
    )
    dispense_flow_rate = (
        sample_transfer_props.dispense.flow_rate_by_volume.get_for_volume(40)
    )
    decoy.when(mock_instrument_core.get_current_volume()).then_return(0, 40)
    aspirate_correction_volume = (
        sample_transfer_props.aspirate.correction_by_volume.get_for_volume(40)
    )
    dispense_correction_volume = (
        sample_transfer_props.dispense.correction_by_volume.get_for_volume(0)
    )
    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=Location(Point(1, 2, 3), labware=None),
        target_well=source_well,
        tip_state=TipState(),
        transfer_type=TransferType.ONE_TO_ONE,
    )
    subject.pre_wet(volume=40)

    decoy.verify(
        mock_instrument_core.aspirate(
            location=Location(Point(1, 2, 3), labware=None),
            well_core=None,
            volume=40,
            rate=1,
            flow_rate=aspirate_flow_rate,
            in_place=True,
            correction_volume=aspirate_correction_volume,
        ),
        mock_instrument_core.delay(0.2),
        mock_instrument_core.dispense(
            location=Location(Point(1, 2, 3), labware=None),
            well_core=None,
            volume=40,
            rate=1,
            flow_rate=dispense_flow_rate,
            in_place=True,
            push_out=0,
            correction_volume=dispense_correction_volume,
        ),
        mock_instrument_core.delay(0.5),
    )


def test_pre_wet_disabled(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
) -> None:
    """It should NOT execute pre-wet steps."""
    source_well = decoy.mock(cls=WellCore)
    sample_transfer_props.aspirate.pre_wet = False
    aspirate_flow_rate = (
        sample_transfer_props.aspirate.flow_rate_by_volume.get_for_volume(40)
    )
    aspirate_correction_volume = (
        sample_transfer_props.aspirate.correction_by_volume.get_for_volume(50)
    )
    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=Location(Point(1, 2, 3), labware=None),
        target_well=source_well,
        tip_state=TipState(),
        transfer_type=TransferType.ONE_TO_ONE,
    )
    subject.pre_wet(volume=40)

    decoy.verify(
        mock_instrument_core.aspirate(
            location=Location(Point(1, 2, 3), labware=None),
            well_core=None,
            volume=40,
            rate=1,
            flow_rate=aspirate_flow_rate,
            in_place=True,
            correction_volume=aspirate_correction_volume,
        ),
        times=0,
    )


@pytest.mark.parametrize(
    argnames=[
        "air_gap_volume",
        "air_gap_flow_rate_by_vol",
        "expected_air_gap_flow_rate",
    ],
    argvalues=[(0.123, 123, 123), (1.23, 0.123, 1.23)],
)
def test_retract_after_aspiration(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
    air_gap_volume: float,
    air_gap_flow_rate_by_vol: float,
    expected_air_gap_flow_rate: float,
) -> None:
    """It should execute steps to retract from well after an aspiration."""
    source_well = decoy.mock(cls=WellCore)
    well_top_point = Point(1, 2, 3)
    well_bottom_point = Point(4, 5, 6)
    air_gap_correction_by_vol = 0.321

    sample_transfer_props.aspirate.retract.air_gap_by_volume.set_for_volume(
        40, air_gap_volume
    )
    sample_transfer_props.aspirate.flow_rate_by_volume.set_for_volume(
        air_gap_volume, air_gap_flow_rate_by_vol
    )
    sample_transfer_props.aspirate.correction_by_volume.set_for_volume(
        air_gap_volume, air_gap_correction_by_vol
    )
    decoy.when(mock_instrument_core.get_current_volume()).then_return(0)

    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=Location(Point(1, 1, 1), labware=None),
        target_well=source_well,
        tip_state=TipState(),
        transfer_type=TransferType.ONE_TO_ONE,
    )
    decoy.when(source_well.get_bottom(0)).then_return(well_bottom_point)
    decoy.when(source_well.get_top(0)).then_return(well_top_point)
    decoy.when(source_well.get_top(AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP)).then_return(
        Point(1, 2, 4)
    )  # Retract location is at same height as safe location
    subject.retract_after_aspiration(volume=40)

    decoy.verify(
        tx_utils.raise_if_location_inside_liquid(
            location=Location(Point(x=4, y=4, z=4), labware=None),
            well_core=source_well,
            location_check_descriptors=LocationCheckDescriptors(
                location_type="retract end",
                pipetting_action="aspirate",
            ),
            logger=matchers.Anything(),
        ),
        mock_instrument_core.move_to(
            location=Location(Point(x=4, y=4, z=4), labware=None),
            well_core=source_well,
            force_direct=True,
            minimum_z_height=None,
            speed=50,
        ),
        mock_instrument_core.delay(20),
        mock_instrument_core.touch_tip(
            location=Location(Point(x=4, y=4, z=4), labware=None),
            well_core=source_well,
            radius=1,
            mm_from_edge=0.5,
            z_offset=-1,
            speed=30,
        ),
        mock_instrument_core.move_to(
            location=Location(Point(x=4, y=4, z=4), labware=None),
            well_core=source_well,
            force_direct=True,
            minimum_z_height=None,
            speed=None,
        ),
        mock_instrument_core.air_gap_in_place(
            volume=air_gap_volume,
            flow_rate=expected_air_gap_flow_rate,
            correction_volume=air_gap_correction_by_vol,
        ),
        mock_instrument_core.delay(0.2),
    )


def test_retract_after_aspiration_when_retract_loc_below_safe_airgap_point(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
) -> None:
    """It should execute post-aspiration retract steps and air gap above retract location."""
    source_well = decoy.mock(cls=WellCore)
    well_top_point = Point(1, 2, 3)
    well_bottom_point = Point(4, 5, 6)
    air_gap_correction_by_vol = 0.321
    air_gap_volume = 0.123
    air_gap_flow_rate_by_vol = 123
    expected_air_gap_flow_rate = 123

    sample_transfer_props.aspirate.retract.air_gap_by_volume.set_for_volume(
        40, air_gap_volume
    )
    sample_transfer_props.aspirate.flow_rate_by_volume.set_for_volume(
        air_gap_volume, air_gap_flow_rate_by_vol
    )
    sample_transfer_props.aspirate.correction_by_volume.set_for_volume(
        air_gap_volume, air_gap_correction_by_vol
    )

    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=Location(Point(1, 1, 1), labware=None),
        target_well=source_well,
        tip_state=TipState(),
        transfer_type=TransferType.ONE_TO_ONE,
    )
    decoy.when(mock_instrument_core.get_current_volume()).then_return(0)
    decoy.when(source_well.get_bottom(0)).then_return(well_bottom_point)
    decoy.when(source_well.get_top(0)).then_return(well_top_point)
    decoy.when(source_well.get_top(AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP)).then_return(
        Point(1, 2, 5)
    )  # Safe location above retract location
    subject.retract_after_aspiration(volume=40)

    decoy.verify(
        tx_utils.raise_if_location_inside_liquid(
            location=Location(Point(x=4, y=4, z=4), labware=None),
            well_core=source_well,
            location_check_descriptors=LocationCheckDescriptors(
                location_type="retract end",
                pipetting_action="aspirate",
            ),
            logger=matchers.Anything(),
        ),
        mock_instrument_core.move_to(
            location=Location(Point(x=4, y=4, z=4), labware=None),
            well_core=source_well,
            force_direct=True,
            minimum_z_height=None,
            speed=50,
        ),
        mock_instrument_core.delay(20),
        mock_instrument_core.touch_tip(
            location=Location(Point(x=4, y=4, z=4), labware=None),
            well_core=source_well,
            radius=1,
            mm_from_edge=0.5,
            z_offset=-1,
            speed=30,
        ),
        mock_instrument_core.move_to(
            location=Location(Point(x=4, y=4, z=4), labware=None),
            well_core=source_well,
            force_direct=True,
            minimum_z_height=None,
            speed=None,
        ),
        mock_instrument_core.move_to(
            location=Location(Point(x=4, y=4, z=5), labware=None),
            well_core=source_well,
            force_direct=True,
            minimum_z_height=None,
            speed=None,
        ),
        mock_instrument_core.air_gap_in_place(
            volume=air_gap_volume,
            flow_rate=expected_air_gap_flow_rate,
            correction_volume=air_gap_correction_by_vol,
        ),
        mock_instrument_core.delay(0.2),
    )


def test_post_aspirate_retract_raises_when_retract_point_is_invalid(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
) -> None:
    """Should raise an error when the retract point is deemed bad."""
    source_well = decoy.mock(cls=WellCore)
    well_top_point = Point(1, 2, 3)
    well_bottom_point = Point(4, 5, 6)
    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=Location(Point(1, 1, 1), labware=None),
        target_well=source_well,
        tip_state=TipState(),
        transfer_type=TransferType.ONE_TO_ONE,
    )
    decoy.when(source_well.get_bottom(0)).then_return(well_bottom_point)
    decoy.when(source_well.get_top(0)).then_return(well_top_point)
    decoy.when(
        tx_utils.raise_if_location_inside_liquid(
            location=Location(Point(x=4, y=4, z=4), labware=None),
            well_core=source_well,
            location_check_descriptors=LocationCheckDescriptors(
                location_type="retract end",
                pipetting_action="aspirate",
            ),
            logger=matchers.Anything(),
        )
    ).then_raise(RuntimeError("Oh no!"))
    with pytest.raises(RuntimeError, match="Oh no!"):
        subject.retract_after_aspiration(volume=40)


def test_retract_after_aspiration_without_touch_tip_and_delay(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
) -> None:
    """It should execute steps to retract from well after an aspiration without a touch tip or delay."""
    source_well = decoy.mock(cls=WellCore)
    well_top_point = Point(1, 2, 3)
    well_bottom_point = Point(4, 5, 6)
    air_gap_volume = 0.123
    air_gap_flow_rate_by_vol = 123
    air_gap_correction_by_vol = 0.321

    sample_transfer_props.aspirate.retract.air_gap_by_volume.set_for_volume(
        40, air_gap_volume
    )
    sample_transfer_props.aspirate.flow_rate_by_volume.set_for_volume(
        air_gap_volume, air_gap_flow_rate_by_vol
    )
    sample_transfer_props.aspirate.correction_by_volume.set_for_volume(
        air_gap_volume, air_gap_correction_by_vol
    )

    sample_transfer_props.aspirate.retract.touch_tip.enabled = False
    sample_transfer_props.aspirate.retract.delay.enabled = False

    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=Location(Point(1, 1, 1), labware=None),
        target_well=source_well,
        tip_state=TipState(
            ready_to_aspirate=True,
            last_liquid_and_air_gap_in_tip=LiquidAndAirGapPair(
                liquid=10,
                air_gap=0,
            ),
        ),
        transfer_type=TransferType.ONE_TO_ONE,
    )
    decoy.when(mock_instrument_core.get_current_volume()).then_return(0)
    decoy.when(source_well.get_bottom(0)).then_return(well_bottom_point)
    decoy.when(source_well.get_top(0)).then_return(well_top_point)
    # Assume air gap safe location is below retract location
    decoy.when(source_well.get_top(AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP)).then_return(
        Point(1, 2, 3)
    )
    subject.retract_after_aspiration(volume=40)

    decoy.verify(
        mock_instrument_core.move_to(
            location=Location(Point(x=4, y=4, z=4), labware=None),
            well_core=source_well,
            force_direct=True,
            minimum_z_height=None,
            speed=50,
        ),
        mock_instrument_core.air_gap_in_place(
            volume=air_gap_volume,
            flow_rate=air_gap_flow_rate_by_vol,
            correction_volume=air_gap_correction_by_vol,
        ),
        mock_instrument_core.delay(0.2),
    )


def test_retract_after_aspiration_for_consolidate(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
) -> None:
    """It should execute steps to retract from well after an aspiration during a MANY_TO_ONE transfer."""
    source_well = decoy.mock(cls=WellCore)
    well_top_point = Point(1, 2, 3)
    well_bottom_point = Point(4, 5, 6)
    air_gap_volume = 0.123
    air_gap_flow_rate_by_vol = 123
    air_gap_correction_by_vol = 0.321

    decoy.when(mock_instrument_core.get_current_volume()).then_return(12.3)
    sample_transfer_props.aspirate.retract.air_gap_by_volume.set_for_volume(
        12.3, air_gap_volume
    )
    sample_transfer_props.aspirate.flow_rate_by_volume.set_for_volume(
        air_gap_volume, air_gap_flow_rate_by_vol
    )
    sample_transfer_props.aspirate.correction_by_volume.set_for_volume(
        12.3 + air_gap_volume, air_gap_correction_by_vol
    )

    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=Location(Point(1, 1, 1), labware=None),
        target_well=source_well,
        tip_state=TipState(),
        transfer_type=TransferType.MANY_TO_ONE,
    )
    decoy.when(source_well.get_bottom(0)).then_return(well_bottom_point)
    decoy.when(source_well.get_top(0)).then_return(well_top_point)
    # Assume air gap safe location is below retract location
    decoy.when(source_well.get_top(AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP)).then_return(
        Point(1, 2, 3)
    )
    subject.retract_after_aspiration(volume=40)

    decoy.verify(
        mock_instrument_core.move_to(
            location=Location(Point(x=4, y=4, z=4), labware=None),
            well_core=source_well,
            force_direct=True,
            minimum_z_height=None,
            speed=50,
        ),
        mock_instrument_core.delay(20),
        mock_instrument_core.touch_tip(
            location=Location(Point(x=4, y=4, z=4), labware=None),
            well_core=source_well,
            radius=1,
            mm_from_edge=0.5,
            z_offset=-1,
            speed=30,
        ),
        mock_instrument_core.move_to(
            location=Location(Point(x=4, y=4, z=4), labware=None),
            well_core=source_well,
            force_direct=True,
            minimum_z_height=None,
            speed=None,
        ),
        mock_instrument_core.air_gap_in_place(
            volume=air_gap_volume,
            flow_rate=air_gap_flow_rate_by_vol,
            correction_volume=air_gap_correction_by_vol,
        ),
        mock_instrument_core.delay(0.2),
    )


"""
Single dispense properties:

"singleDispense": {
    "submerge": {
      "startPosition": {"positionReference": "well-top", "offset": {"x": 30, "y": 20, "z": 10}},
      "speed": 100,
      "delay": {"enable": true, "params": { "duration": 0.0 }}
    },
    "retract": {
      "endPosition": {"positionReference": "well-top", "offset": {"x": 11, "y": 22, "z": 33}},
      "speed": 50,
      "airGapByVolume": [[1.0, 0.1], [49.9, 0.1], [50.0, 0.0]],
      "blowout": { "enable": true , "params": {"location": "source", "flowRate": 100}},
      "touchTip": { "enable": true, "params": { "zOffset": -1, "mmFromEdge": 0.5, "speed": 30}},
      "delay": {"enable": true, "params": { "duration": 10 }}
    },
    "dispensePosition": {"positionReference": "well-bottom", "offset": {"x": 33, "y": 22, "z": 11}},
    "flowRateByVolume": [[1.0, 50.0]],
    "correctionByVolume": [[0.0, 0.0]],
    "mix": { "enable": true, "params": { "repetitions": 1, "volume": 50 }},
    "pushOutByVolume": [[1.0, 7.0], [4.999, 7.0], [5.0, 2.0], [10.0, 2.0], [50.0, 2.0]],
    "delay": { "enable": true, "params": { "duration": 0.2 }},
}
"""


@pytest.mark.parametrize(
    "add_final_air_gap",
    [True, False],
)
def test_retract_after_dispense_with_blowout_in_source(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
    add_final_air_gap: bool,
) -> None:
    """It should execute steps to retract from well after a dispense."""
    source_location = Location(Point(1, 2, 3), labware=None)
    source_well = decoy.mock(cls=WellCore)
    dest_well = decoy.mock(cls=WellCore)
    well_top_point = Point(1, 2, 3)
    well_bottom_point = Point(4, 5, 6)
    air_gap_volume = 0.123
    air_gap_flow_rate_by_vol = 123
    air_gap_correction_by_vol = 0.321

    sample_transfer_props.dispense.retract.air_gap_by_volume.set_for_volume(
        0, air_gap_volume
    )
    sample_transfer_props.aspirate.flow_rate_by_volume.set_for_volume(
        air_gap_volume, air_gap_flow_rate_by_vol
    )
    sample_transfer_props.aspirate.correction_by_volume.set_for_volume(
        air_gap_volume, air_gap_correction_by_vol
    )

    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=Location(Point(1, 1, 1), labware=None),
        target_well=dest_well,
        tip_state=TipState(),
        transfer_type=TransferType.ONE_TO_ONE,
    )
    decoy.when(mock_instrument_core.get_current_volume()).then_return(0)
    decoy.when(dest_well.get_bottom(0)).then_return(well_bottom_point)
    decoy.when(dest_well.get_top(0)).then_return(well_top_point)
    decoy.when(source_well.get_top(0)).then_return(Point(10, 20, 30))
    # Assume air gap safe location is below retract location
    decoy.when(dest_well.get_top(AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP)).then_return(
        Point(1, 2, 35)
    )
    # Assume air gap safe location for air-gapping at src is below blowout position,
    # where blowout position is source well top
    decoy.when(source_well.get_top(AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP)).then_return(
        Point(10, 20, 29)
    )
    subject.retract_after_dispensing(
        trash_location=Location(Point(), labware=None),
        source_location=source_location,
        source_well=source_well,
        add_final_air_gap=add_final_air_gap,
    )
    decoy.verify(
        mock_instrument_core.move_to(
            location=Location(Point(12, 24, 36), labware=None),
            well_core=dest_well,
            force_direct=True,
            minimum_z_height=None,
            speed=50,
        ),
        mock_instrument_core.delay(10),
        mock_instrument_core.touch_tip(
            location=Location(Point(12, 24, 36), labware=None),
            well_core=dest_well,
            radius=1,
            mm_from_edge=0.75,
            z_offset=-1,
            speed=30,
        ),
        mock_instrument_core.move_to(
            location=Location(Point(12, 24, 36), labware=None),
            well_core=dest_well,
            force_direct=True,
            minimum_z_height=None,
            speed=None,
        ),
        mock_instrument_core.air_gap_in_place(
            volume=air_gap_volume,
            flow_rate=air_gap_flow_rate_by_vol,
            correction_volume=air_gap_correction_by_vol,
        ),
        mock_instrument_core.delay(0.2),
        mock_instrument_core.blow_out(
            location=Location(Point(10, 20, 30), labware=None),
            well_core=source_well,
            in_place=False,
            flow_rate=100,
        ),
        mock_instrument_core.touch_tip(
            location=Location(Point(10, 20, 30), labware=None),
            well_core=source_well,
            radius=1,
            mm_from_edge=0.75,
            z_offset=-1,
            speed=30,
        ),
        mock_instrument_core.move_to(
            location=Location(Point(10, 20, 30), labware=None),
            well_core=source_well,
            force_direct=True,
            minimum_z_height=None,
            speed=None,
        ),
        mock_instrument_core.prepare_to_aspirate(),
        *(
            add_final_air_gap
            and [
                mock_instrument_core.air_gap_in_place(  # type: ignore[func-returns-value]
                    volume=air_gap_volume,
                    flow_rate=air_gap_flow_rate_by_vol,
                    correction_volume=air_gap_correction_by_vol,
                ),
                mock_instrument_core.delay(0.2),  # type: ignore[func-returns-value]
            ]
            or []
        ),
    )


@pytest.mark.parametrize(
    "add_final_air_gap",
    [True, False],
)
def test_retract_after_dispense_with_blowout_in_destination(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
    add_final_air_gap: bool,
) -> None:
    """It should execute steps to retract from well after a dispense."""
    source_well = decoy.mock(cls=WellCore)
    dest_well = decoy.mock(cls=WellCore)
    well_top_point = Point(1, 2, 3)
    well_bottom_point = Point(4, 5, 6)
    air_gap_volume = 0.123
    air_gap_flow_rate_by_vol = 123
    air_gap_correction_by_vol = 0.321

    sample_transfer_props.aspirate.retract.air_gap_by_volume.set_for_volume(
        0,
        1234,  # Explicitly check that this value is not used during post-dispense air gap
    )
    sample_transfer_props.dispense.retract.air_gap_by_volume.set_for_volume(
        0, air_gap_volume
    )
    sample_transfer_props.aspirate.flow_rate_by_volume.set_for_volume(
        air_gap_volume, air_gap_flow_rate_by_vol
    )
    sample_transfer_props.aspirate.correction_by_volume.set_for_volume(
        air_gap_volume, air_gap_correction_by_vol
    )

    sample_transfer_props.dispense.retract.blowout.location = cast(
        str, BlowoutLocation.DESTINATION
    )

    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=Location(Point(1, 1, 1), labware=None),
        target_well=dest_well,
        tip_state=TipState(
            ready_to_aspirate=True,
            last_liquid_and_air_gap_in_tip=LiquidAndAirGapPair(
                liquid=10,
                air_gap=0,
            ),
        ),
        transfer_type=TransferType.ONE_TO_ONE,
    )
    decoy.when(mock_instrument_core.get_current_volume()).then_return(0)
    decoy.when(dest_well.get_bottom(0)).then_return(well_bottom_point)
    decoy.when(dest_well.get_top(0)).then_return(well_top_point)
    # Assume air gap safe location is below retract location
    decoy.when(dest_well.get_top(AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP)).then_return(
        Point(1, 2, 35)
    )
    subject.retract_after_dispensing(
        trash_location=Location(Point(), labware=None),
        source_location=Location(Point(1, 2, 3), labware=None),
        source_well=source_well,
        add_final_air_gap=True,
    )
    decoy.verify(
        mock_instrument_core.move_to(
            location=Location(Point(12, 24, 36), labware=None),
            well_core=dest_well,
            force_direct=True,
            minimum_z_height=None,
            speed=50,
        ),
        mock_instrument_core.delay(10),
        mock_instrument_core.blow_out(
            location=Location(Point(12, 24, 36), labware=None),
            well_core=dest_well,
            in_place=True,
            flow_rate=100,
        ),
        mock_instrument_core.touch_tip(
            location=Location(Point(12, 24, 36), labware=None),
            well_core=dest_well,
            radius=1,
            mm_from_edge=0.75,
            z_offset=-1,
            speed=30,
        ),
        mock_instrument_core.move_to(
            location=Location(Point(12, 24, 36), labware=None),
            well_core=dest_well,
            force_direct=True,
            minimum_z_height=None,
            speed=None,
        ),
        *(
            add_final_air_gap
            and [
                mock_instrument_core.air_gap_in_place(  # type: ignore[func-returns-value]
                    volume=air_gap_volume,
                    flow_rate=air_gap_flow_rate_by_vol,
                    correction_volume=air_gap_correction_by_vol,
                ),
                mock_instrument_core.delay(0.2),  # type: ignore[func-returns-value]
            ]
            or []
        ),
    )


@pytest.mark.parametrize(
    "add_final_air_gap",
    [True, False],
)
def test_retract_after_dispense_with_blowout_in_trash_well(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
    add_final_air_gap: bool,
) -> None:
    """It should execute steps to retract from well after a dispense."""
    source_location = Location(Point(1, 2, 3), labware=None)
    source_well = decoy.mock(cls=WellCore)
    dest_well = decoy.mock(cls=WellCore)
    trash_well = decoy.mock(cls=Well)
    trash_well_core = decoy.mock(cls=WellCore)
    trash_location = Location(Point(7, 8, 9), labware=trash_well)
    well_top_point = Point(1, 2, 3)
    well_bottom_point = Point(4, 5, 6)
    air_gap_volume = 0.123
    air_gap_flow_rate_by_vol = 123
    air_gap_correction_by_vol = 0.321

    sample_transfer_props.dispense.retract.air_gap_by_volume.set_for_volume(
        0, air_gap_volume
    )
    sample_transfer_props.aspirate.flow_rate_by_volume.set_for_volume(
        air_gap_volume, air_gap_flow_rate_by_vol
    )
    sample_transfer_props.aspirate.correction_by_volume.set_for_volume(
        air_gap_volume, air_gap_correction_by_vol
    )

    sample_transfer_props.dispense.retract.blowout.location = cast(
        str, BlowoutLocation.TRASH
    )

    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=Location(Point(1, 1, 1), labware=None),
        target_well=dest_well,
        tip_state=TipState(),
        transfer_type=TransferType.ONE_TO_ONE,
    )
    decoy.when(mock_instrument_core.get_current_volume()).then_return(0)
    decoy.when(dest_well.get_bottom(0)).then_return(well_bottom_point)
    decoy.when(dest_well.get_top(0)).then_return(well_top_point)
    decoy.when(trash_well._core).then_return(trash_well_core)
    # Assume air gap safe location is below retract location
    decoy.when(dest_well.get_top(AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP)).then_return(
        Point(1, 2, 35)
    )
    # Assume air gap safe location for air-gapping at src is below touch-tip position,
    # where touch tip position is source well top
    decoy.when(trash_well_core.get_top(AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP)).then_return(
        Point(7, 8, 8)
    )
    subject.retract_after_dispensing(
        trash_location=trash_location,
        source_location=source_location,
        source_well=source_well,
        add_final_air_gap=True,
    )
    decoy.verify(
        mock_instrument_core.move_to(
            location=Location(Point(12, 24, 36), labware=None),
            well_core=dest_well,
            force_direct=True,
            minimum_z_height=None,
            speed=50,
        ),
        mock_instrument_core.delay(10),
        mock_instrument_core.touch_tip(
            location=Location(Point(12, 24, 36), labware=None),
            well_core=dest_well,
            radius=1,
            mm_from_edge=0.75,
            z_offset=-1,
            speed=30,
        ),
        mock_instrument_core.move_to(
            location=Location(Point(12, 24, 36), labware=None),
            well_core=dest_well,
            force_direct=True,
            minimum_z_height=None,
            speed=None,
        ),
        mock_instrument_core.air_gap_in_place(
            volume=air_gap_volume,
            flow_rate=air_gap_flow_rate_by_vol,
            correction_volume=air_gap_correction_by_vol,
        ),
        mock_instrument_core.delay(0.2),
        mock_instrument_core.blow_out(
            location=trash_location,
            well_core=trash_well_core,
            in_place=False,
            flow_rate=100,
        ),
        mock_instrument_core.touch_tip(
            location=trash_location,
            well_core=trash_well_core,
            radius=1,
            mm_from_edge=0.75,
            z_offset=-1,
            speed=30,
        ),
        mock_instrument_core.move_to(
            location=trash_location,
            well_core=trash_well_core,
            force_direct=True,
            minimum_z_height=None,
            speed=None,
        ),
        *(
            add_final_air_gap
            and [
                mock_instrument_core.air_gap_in_place(  # type: ignore[func-returns-value]
                    volume=air_gap_volume,
                    flow_rate=air_gap_flow_rate_by_vol,
                    correction_volume=air_gap_correction_by_vol,
                ),
                mock_instrument_core.delay(0.2),  # type: ignore[func-returns-value]
            ]
            or []
        ),
    )


@pytest.mark.parametrize(
    "add_final_air_gap",
    [True, False],
)
def test_retract_after_dispense_with_blowout_in_disposal_location(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
    add_final_air_gap: bool,
) -> None:
    """It should execute steps to retract from well after a dispense."""
    source_location = Location(Point(1, 2, 3), labware=None)
    source_well = decoy.mock(cls=WellCore)
    dest_well = decoy.mock(cls=WellCore)
    trash_location = decoy.mock(cls=TrashBin)
    trash_top = decoy.mock(cls=TrashBin)
    well_top_point = Point(1, 2, 3)
    well_bottom_point = Point(4, 5, 6)
    air_gap_volume = 0.123
    air_gap_flow_rate_by_vol = 123
    air_gap_correction_by_vol = 0.321

    sample_transfer_props.dispense.retract.air_gap_by_volume.set_for_volume(
        0, air_gap_volume
    )
    sample_transfer_props.aspirate.flow_rate_by_volume.set_for_volume(
        air_gap_volume, air_gap_flow_rate_by_vol
    )
    sample_transfer_props.aspirate.correction_by_volume.set_for_volume(
        air_gap_volume, air_gap_correction_by_vol
    )

    sample_transfer_props.dispense.retract.blowout.location = cast(
        str, BlowoutLocation.TRASH
    )

    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=Location(Point(1, 1, 1), labware=None),
        target_well=dest_well,
        tip_state=TipState(),
        transfer_type=TransferType.ONE_TO_ONE,
    )
    decoy.when(mock_instrument_core.get_current_volume()).then_return(0)
    decoy.when(dest_well.get_bottom(0)).then_return(well_bottom_point)
    decoy.when(dest_well.get_top(0)).then_return(well_top_point)
    # Assume air gap safe location is below retract location
    decoy.when(dest_well.get_top(AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP)).then_return(
        Point(1, 2, 35)
    )
    decoy.when(trash_location.offset).then_return(DisposalOffset(x=4, y=5, z=6))
    decoy.when(trash_location.top(x=0, y=0, z=2)).then_return(trash_top)
    decoy.when(trash_top.offset).then_return(DisposalOffset(x=1, y=2, z=3))
    subject.retract_after_dispensing(
        trash_location=trash_location,
        source_location=source_location,
        source_well=source_well,
        add_final_air_gap=True,
    )
    decoy.verify(
        mock_instrument_core.move_to(
            location=Location(Point(12, 24, 36), labware=None),
            well_core=dest_well,
            force_direct=True,
            minimum_z_height=None,
            speed=50,
        ),
        mock_instrument_core.delay(10),
        mock_instrument_core.touch_tip(
            location=Location(Point(12, 24, 36), labware=None),
            well_core=dest_well,
            radius=1,
            mm_from_edge=0.75,
            z_offset=-1,
            speed=30,
        ),
        mock_instrument_core.move_to(
            location=Location(Point(12, 24, 36), labware=None),
            well_core=dest_well,
            force_direct=True,
            minimum_z_height=None,
            speed=None,
        ),
        mock_instrument_core.air_gap_in_place(
            volume=air_gap_volume,
            flow_rate=air_gap_flow_rate_by_vol,
            correction_volume=air_gap_correction_by_vol,
        ),
        mock_instrument_core.delay(0.2),
        mock_instrument_core.blow_out(
            location=trash_location,
            well_core=None,
            in_place=False,
            flow_rate=100,
        ),
        *(
            add_final_air_gap
            and [
                mock_instrument_core.air_gap_in_place(  # type: ignore[func-returns-value]
                    volume=air_gap_volume,
                    flow_rate=air_gap_flow_rate_by_vol,
                    correction_volume=air_gap_correction_by_vol,
                ),
                mock_instrument_core.delay(0.2),  # type: ignore[func-returns-value]
            ]
            or []
        ),
    )


@pytest.mark.parametrize(
    "add_final_air_gap",
    [True, False],
)
def test_retract_after_dispense_in_trash_with_blowout_in_source(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
    add_final_air_gap: bool,
) -> None:
    """It should execute steps to retract after a dispense into a trash."""
    source_location = Location(Point(1, 2, 3), labware=None)
    source_well = decoy.mock(cls=WellCore)
    target_chute = decoy.mock(cls=WasteChute)
    chute_top = decoy.mock(cls=WasteChute)
    air_gap_volume = 0.123
    air_gap_flow_rate_by_vol = 123
    air_gap_correction_by_vol = 0.321

    sample_transfer_props.dispense.retract.air_gap_by_volume.set_for_volume(
        0, air_gap_volume
    )
    sample_transfer_props.aspirate.flow_rate_by_volume.set_for_volume(
        air_gap_volume, air_gap_flow_rate_by_vol
    )
    sample_transfer_props.aspirate.correction_by_volume.set_for_volume(
        air_gap_volume, air_gap_correction_by_vol
    )

    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=target_chute,
        target_well=None,
        tip_state=TipState(),
        transfer_type=TransferType.ONE_TO_ONE,
    )

    decoy.when(mock_instrument_core.get_current_volume()).then_return(0)
    decoy.when(source_well.get_top(0)).then_return(Point(10, 20, 30))
    # Assume air gap safe location for air-gapping at src is below touch-tip position,
    # where touch tip position is source well top
    decoy.when(source_well.get_top(AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP)).then_return(
        Point(10, 20, 29)
    )
    decoy.when(target_chute.offset).then_return(DisposalOffset(x=4, y=5, z=6))
    decoy.when(target_chute.top(x=0, y=0, z=2)).then_return(chute_top)
    decoy.when(chute_top.offset).then_return(DisposalOffset(x=1, y=2, z=3))
    subject.retract_after_dispensing(
        trash_location=Location(Point(), labware=None),
        source_location=source_location,
        source_well=source_well,
        add_final_air_gap=add_final_air_gap,
    )
    decoy.verify(
        mock_instrument_core.delay(10),
        mock_instrument_core.air_gap_in_place(
            volume=air_gap_volume,
            flow_rate=air_gap_flow_rate_by_vol,
            correction_volume=air_gap_correction_by_vol,
        ),
        mock_instrument_core.delay(0.2),
        mock_instrument_core.blow_out(
            location=Location(Point(10, 20, 30), labware=None),
            well_core=source_well,
            in_place=False,
            flow_rate=100,
        ),
        mock_instrument_core.touch_tip(
            location=Location(Point(10, 20, 30), labware=None),
            well_core=source_well,
            radius=1,
            mm_from_edge=0.75,
            z_offset=-1,
            speed=30,
        ),
        mock_instrument_core.move_to(
            location=Location(Point(10, 20, 30), labware=None),
            well_core=source_well,
            force_direct=True,
            minimum_z_height=None,
            speed=None,
        ),
        mock_instrument_core.prepare_to_aspirate(),
        *(
            add_final_air_gap
            and [
                mock_instrument_core.air_gap_in_place(  # type: ignore[func-returns-value]
                    volume=air_gap_volume,
                    flow_rate=air_gap_flow_rate_by_vol,
                    correction_volume=air_gap_correction_by_vol,
                ),
                mock_instrument_core.delay(0.2),  # type: ignore[func-returns-value]
            ]
            or []
        ),
    )


@pytest.mark.parametrize(
    "add_final_air_gap",
    [True, False],
)
def test_retract_after_dispense_in_trash_with_blowout_in_destination(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
    add_final_air_gap: bool,
) -> None:
    """It should execute steps to retract after a dispense into a trash."""
    source_well = decoy.mock(cls=WellCore)
    target_trash = decoy.mock(cls=TrashBin)
    trash_top = decoy.mock(cls=TrashBin)
    air_gap_volume = 0.123
    air_gap_flow_rate_by_vol = 123
    air_gap_correction_by_vol = 0.321

    sample_transfer_props.dispense.retract.air_gap_by_volume.set_for_volume(
        0, air_gap_volume
    )
    sample_transfer_props.aspirate.flow_rate_by_volume.set_for_volume(
        air_gap_volume, air_gap_flow_rate_by_vol
    )
    sample_transfer_props.aspirate.correction_by_volume.set_for_volume(
        air_gap_volume, air_gap_correction_by_vol
    )

    sample_transfer_props.dispense.retract.blowout.location = cast(
        str, BlowoutLocation.DESTINATION
    )

    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=target_trash,
        target_well=None,
        tip_state=TipState(
            ready_to_aspirate=True,
            last_liquid_and_air_gap_in_tip=LiquidAndAirGapPair(
                liquid=10,
                air_gap=0,
            ),
        ),
        transfer_type=TransferType.ONE_TO_ONE,
    )
    decoy.when(mock_instrument_core.get_current_volume()).then_return(0)
    decoy.when(target_trash.offset).then_return(DisposalOffset(x=4, y=5, z=6))
    decoy.when(target_trash.top(x=0, y=0, z=2)).then_return(trash_top)
    decoy.when(trash_top.offset).then_return(DisposalOffset(x=1, y=2, z=3))
    subject.retract_after_dispensing(
        trash_location=Location(Point(), labware=None),
        source_location=Location(Point(1, 2, 3), labware=None),
        source_well=source_well,
        add_final_air_gap=True,
    )
    decoy.verify(
        mock_instrument_core.delay(10),
        mock_instrument_core.blow_out(
            location=target_trash,
            well_core=None,
            in_place=True,
            flow_rate=100,
        ),
        *(
            add_final_air_gap
            and [
                mock_instrument_core.air_gap_in_place(  # type: ignore[func-returns-value]
                    volume=air_gap_volume,
                    flow_rate=air_gap_flow_rate_by_vol,
                    correction_volume=air_gap_correction_by_vol,
                ),
                mock_instrument_core.delay(0.2),  # type: ignore[func-returns-value]
            ]
            or []
        ),
    )


@pytest.mark.parametrize(
    "add_final_air_gap",
    [True, False],
)
def test_retract_after_dispense_in_trash_with_blowout_in_disposal_location(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
    add_final_air_gap: bool,
) -> None:
    """It should execute steps to retract after a dispense into a trash."""
    source_location = Location(Point(1, 2, 3), labware=None)
    source_well = decoy.mock(cls=WellCore)
    target_trash = decoy.mock(cls=TrashBin)
    target_trash_top = decoy.mock(cls=TrashBin)
    trash_location = decoy.mock(cls=WasteChute)
    waste_chute_top = decoy.mock(cls=WasteChute)

    air_gap_volume = 0.123
    air_gap_flow_rate_by_vol = 123
    air_gap_correction_by_vol = 0.321

    sample_transfer_props.dispense.retract.air_gap_by_volume.set_for_volume(
        0, air_gap_volume
    )
    sample_transfer_props.aspirate.flow_rate_by_volume.set_for_volume(
        air_gap_volume, air_gap_flow_rate_by_vol
    )
    sample_transfer_props.aspirate.correction_by_volume.set_for_volume(
        air_gap_volume, air_gap_correction_by_vol
    )

    sample_transfer_props.dispense.retract.blowout.location = cast(
        str, BlowoutLocation.TRASH
    )

    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=target_trash,
        target_well=None,
        tip_state=TipState(),
        transfer_type=TransferType.ONE_TO_ONE,
    )
    decoy.when(mock_instrument_core.get_current_volume()).then_return(0)
    decoy.when(target_trash.offset).then_return(DisposalOffset(x=4, y=5, z=6))
    decoy.when(target_trash.top(x=0, y=0, z=2)).then_return(target_trash_top)
    decoy.when(target_trash_top.offset).then_return(DisposalOffset(x=1, y=2, z=3))
    decoy.when(trash_location.offset).then_return(DisposalOffset(x=4, y=5, z=6))
    decoy.when(trash_location.top(x=0, y=0, z=2)).then_return(waste_chute_top)
    decoy.when(waste_chute_top.offset).then_return(DisposalOffset(x=1, y=2, z=3))
    subject.retract_after_dispensing(
        trash_location=trash_location,
        source_location=source_location,
        source_well=source_well,
        add_final_air_gap=True,
    )
    decoy.verify(
        mock_instrument_core.delay(10),
        mock_instrument_core.air_gap_in_place(
            volume=air_gap_volume,
            flow_rate=air_gap_flow_rate_by_vol,
            correction_volume=air_gap_correction_by_vol,
        ),
        mock_instrument_core.delay(0.2),
        mock_instrument_core.blow_out(
            location=trash_location,
            well_core=None,
            in_place=False,
            flow_rate=100,
        ),
        *(
            add_final_air_gap
            and [
                mock_instrument_core.air_gap_in_place(  # type: ignore[func-returns-value]
                    volume=air_gap_volume,
                    flow_rate=air_gap_flow_rate_by_vol,
                    correction_volume=air_gap_correction_by_vol,
                ),
                mock_instrument_core.delay(0.2),  # type: ignore[func-returns-value]
            ]
            or []
        ),
    )


@pytest.mark.parametrize(
    argnames=[
        "blowout_location",
        "blowout_position",
        "blowout_position_reference_point",
        "expected_blowout_point",
    ],
    argvalues=[
        (
            "trash",
            {"position_reference": "well-top", "offset": {"x": 1, "y": 2, "z": 3}},
            Point(1, 2, 3),  # Well-top point
            Point(2, 4, 6),
        ),
        (
            "destination",
            {
                "position_reference": "well-center",
                "offset": {"x": 10, "y": 20, "z": 30},
            },
            Point(1, 1, 1),  # Well-center point
            Point(11, 21, 31),
        ),
        (
            "source",
            {
                "position_reference": "well-bottom",
                "offset": {"x": 100, "y": 200, "z": 300},
            },
            Point(0, 0, 0),  # Well-bottom point
            Point(100, 200, 300),
        ),
    ],
)
def test_retract_after_dispense_with_blowout_position_set(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
    blowout_location: str,
    blowout_position: TipPositionDict,
    blowout_position_reference_point: Point,
    expected_blowout_point: Point,
) -> None:
    """It should use the blowout position for destination blowout."""
    source_well_core = decoy.mock(cls=WellCore)
    source_well = decoy.mock(cls=Well)
    source_location = Location(Point(1, 2, 3), labware=source_well)

    trash_well_core = decoy.mock(cls=WellCore)
    trash_well = decoy.mock(cls=Well)
    trash_location = Location(Point(4, 5, 6), labware=trash_well)

    dest_well_core = decoy.mock(cls=WellCore)
    dest_well = decoy.mock(cls=Well)

    blowout_well_core = {
        "trash": trash_well_core,
        "destination": dest_well_core,
        "source": source_well_core,
    }[blowout_location]
    blowout_well = {
        "trash": trash_well,
        "destination": dest_well,
        "source": source_well,
    }[blowout_location]

    well_top_point = Point(1, 2, 3)
    sample_transfer_props.dispense.retract.blowout.location = blowout_location
    sample_transfer_props.dispense.retract.blowout.blowout_position = blowout_position
    # Disable touch tip so we don't have to mock out touch tip related functions
    sample_transfer_props.dispense.retract.touch_tip.enabled = False
    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=Location(Point(1, 1, 1), labware=dest_well),
        target_well=dest_well_core,
        tip_state=TipState(),
        transfer_type=TransferType.ONE_TO_ONE,
    )
    decoy.when(trash_well._core).then_return(trash_well_core)
    decoy.when(mock_instrument_core.get_current_volume()).then_return(0)
    decoy.when(dest_well_core.get_top(0)).then_return(well_top_point)
    decoy.when(source_well_core.get_top(0)).then_return(well_top_point)

    for well_core in [trash_well_core, source_well_core, dest_well_core]:
        decoy.when(well_core.get_top(AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP)).then_return(
            Point(10, 20, 29)
        )
    decoy.when(trash_well_core.get_top(0)).then_return(blowout_position_reference_point)
    decoy.when(dest_well_core.get_center()).then_return(
        blowout_position_reference_point
    )
    decoy.when(source_well_core.get_bottom(0)).then_return(
        blowout_position_reference_point
    )

    subject.retract_after_dispensing(
        trash_location=trash_location,
        source_location=source_location,
        source_well=source_well_core,
        add_final_air_gap=False,
    )
    decoy.verify(
        mock_instrument_core.blow_out(
            location=Location(expected_blowout_point, labware=blowout_well),
            well_core=blowout_well_core,
            in_place=False,
            flow_rate=100,
        )
    )


def test_retract_after_dispense_with_blowout_position_for_trashbin(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
) -> None:
    """It should use the blowout position for destination blowout."""
    trash_location = decoy.mock(cls=TrashBin)
    expected_blowout_trash_location = decoy.mock(
        cls=WasteChute
    )  # Use waste chute just to make the expected result object distinct

    source_well_core = decoy.mock(cls=WellCore)
    source_location = Location(Point(1, 2, 3), labware=decoy.mock(cls=Well))
    dest_well_core = decoy.mock(cls=WellCore)
    dest_location = Location(Point(1, 1, 1), labware=decoy.mock(cls=Well))

    well_top_point = Point(1, 2, 3)
    sample_transfer_props.dispense.retract.blowout.location = "trash"
    sample_transfer_props.dispense.retract.blowout.blowout_position = {
        "position_reference": "well-center",
        "offset": {"x": 10, "y": 20, "z": 30},
    }
    # Disable touch tip so we don't have to mock out touch tip related functions
    sample_transfer_props.dispense.retract.touch_tip.enabled = False
    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=dest_location,
        target_well=dest_well_core,
        tip_state=TipState(),
        transfer_type=TransferType.ONE_TO_ONE,
    )
    decoy.when(mock_instrument_core.get_current_volume()).then_return(0)
    decoy.when(dest_well_core.get_top(0)).then_return(well_top_point)

    for well_core in [source_well_core, dest_well_core]:
        decoy.when(well_core.get_top(AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP)).then_return(
            Point(10, 20, 29)
        )

    decoy.when(
        trash_location.top(x=0, y=0, z=AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP)
    ).then_return(trash_location)
    decoy.when(trash_location.offset).then_return(DisposalOffset(x=10, y=20, z=30))
    decoy.when(
        tx_utils.get_blowout_location_for_trash(
            trash_location,
            sample_transfer_props.dispense.retract.blowout.blowout_position,  # type: ignore[arg-type]
        )
    ).then_return(expected_blowout_trash_location)

    subject.retract_after_dispensing(
        trash_location=trash_location,
        source_location=source_location,
        source_well=source_well_core,
        add_final_air_gap=False,
    )
    decoy.verify(
        mock_instrument_core.blow_out(
            location=expected_blowout_trash_location,
            well_core=None,
            in_place=False,
            flow_rate=100,
        )
    )


def test_retract_after_dispense_with_blowout_in_trash_destination(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
) -> None:
    """It should blowout at the specified position in a trash destination."""
    dest_location = decoy.mock(cls=WasteChute)
    expected_blowout_dest_location = decoy.mock(
        cls=TrashBin
    )  # Use trash bin just to make the expected result object distinct

    source_well_core = decoy.mock(cls=WellCore)
    source_location = Location(Point(1, 2, 3), labware=decoy.mock(cls=Well))

    trash_well_core = decoy.mock(cls=WellCore)
    trash_location = Location(Point(4, 5, 6), labware=decoy.mock(cls=Well))

    sample_transfer_props.dispense.retract.blowout.location = "destination"
    sample_transfer_props.dispense.retract.blowout.blowout_position = {
        "position_reference": "well-center",
        "offset": {"x": 10, "y": 20, "z": 30},
    }
    # Disable touch tip so we don't have to mock out touch tip related functions
    sample_transfer_props.dispense.retract.touch_tip.enabled = False
    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=dest_location,
        target_well=None,
        tip_state=TipState(),
        transfer_type=TransferType.ONE_TO_ONE,
    )
    decoy.when(mock_instrument_core.get_current_volume()).then_return(0)
    for well_core in [source_well_core, trash_well_core]:
        decoy.when(well_core.get_top(AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP)).then_return(
            Point(10, 20, 29)
        )

    decoy.when(
        dest_location.top(x=0, y=0, z=AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP)
    ).then_return(dest_location)
    decoy.when(dest_location.offset).then_return(DisposalOffset(x=10, y=20, z=30))
    decoy.when(
        tx_utils.get_blowout_location_for_trash(
            dest_location,
            sample_transfer_props.dispense.retract.blowout.blowout_position,  # type: ignore[arg-type]
        )
    ).then_return(expected_blowout_dest_location)

    subject.retract_after_dispensing(
        trash_location=trash_location,
        source_location=source_location,
        source_well=source_well_core,
        add_final_air_gap=False,
    )
    decoy.verify(
        mock_instrument_core.blow_out(
            location=expected_blowout_dest_location,
            well_core=None,
            in_place=False,
            flow_rate=100,
        )
    )


# Try all combinations of destination type, blowout location, and touchability
@pytest.mark.parametrize("dest_type", ["well", "trash"])
@pytest.mark.parametrize("blowout_location", [bl for bl in BlowoutLocation])
@pytest.mark.parametrize("source_touchable", [True, False])
def test_retract_after_dispense_touch_tip(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
    dest_type: Literal["well", "trash"],
    blowout_location: BlowoutLocation,
    source_touchable: bool,
) -> None:
    """Should not fail if touchTipDisabled at the blowout location if blowing out to SOURCE."""
    source_location_labware = decoy.mock(cls=Labware)
    source_location = Location(Point(1, 2, 3), labware=source_location_labware)
    source_well = decoy.mock(cls=WellCore)
    dest_well = decoy.mock(cls=WellCore)
    well_top_point = Point(1, 2, 3)
    trash_location = decoy.mock(cls=TrashBin)
    trash_top = decoy.mock(cls=TrashBin)

    target_location: Union[Location, TrashBin, WasteChute]
    if dest_type == "trash":
        target_location = trash_location
        target_well = None
    else:  # dest_type == "well"
        target_location = Location(Point(1, 1, 1), labware=None)
        target_well = dest_well

    sample_transfer_props.dispense.retract.touch_tip.enabled = True
    sample_transfer_props.dispense.retract.blowout.enabled = True
    sample_transfer_props.dispense.retract.blowout.location = cast(
        str, blowout_location
    )

    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=target_location,
        target_well=target_well,
        tip_state=TipState(
            ready_to_aspirate=True,
            last_liquid_and_air_gap_in_tip=LiquidAndAirGapPair(liquid=0, air_gap=0),
        ),
        transfer_type=TransferType.ONE_TO_ONE,
    )

    decoy.when(mock_instrument_core.get_current_volume()).then_return(0)
    decoy.when(source_location_labware.quirks).then_return(
        [] if source_touchable else ["touchTipDisabled"]
    )
    decoy.when(source_well.get_top(0)).then_return(well_top_point)
    decoy.when(source_well.get_top(AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP)).then_return(
        well_top_point + Point(0, 0, AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP)
    )
    decoy.when(dest_well.get_top(0)).then_return(well_top_point)
    decoy.when(dest_well.get_top(AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP)).then_return(
        well_top_point + Point(0, 0, AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP)
    )
    decoy.when(trash_location.offset).then_return(DisposalOffset(x=0, y=0, z=0))
    decoy.when(
        trash_location.top(x=0, y=0, z=AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP)
    ).then_return(trash_top)
    decoy.when(trash_top.offset).then_return(
        DisposalOffset(x=0, y=0, z=AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP)
    )

    subject.retract_after_dispensing(
        trash_location=trash_location,
        source_location=source_location,
        source_well=source_well,
        add_final_air_gap=False,
    )

    decoy.verify(
        mock_instrument_core.touch_tip(
            location=matchers.Anything(),
            well_core=matchers.Anything(),
            radius=matchers.Anything(),
            z_offset=matchers.Anything(),
            speed=matchers.Anything(),
        ),
        ignore_extra_args=True,
        times={
            # (destination type, blowout location, source touchable):
            ("well", BlowoutLocation.SOURCE, True): 2,  # touch dest, touch source
            ("well", BlowoutLocation.SOURCE, False): 1,  # touch dest, no touch source
            ("well", BlowoutLocation.DESTINATION, True): 1,  # touch dest
            ("well", BlowoutLocation.DESTINATION, False): 1,  # touch dest
            ("well", BlowoutLocation.TRASH, True): 1,  # touch dest
            ("well", BlowoutLocation.TRASH, False): 1,  # touch dest
            ("trash", BlowoutLocation.SOURCE, True): 1,  # touch source
            ("trash", BlowoutLocation.SOURCE, False): 0,  # don't touch source
            ("trash", BlowoutLocation.DESTINATION, True): 0,
            ("trash", BlowoutLocation.DESTINATION, False): 0,
            ("trash", BlowoutLocation.TRASH, True): 0,
            ("trash", BlowoutLocation.TRASH, False): 0,
        }[dest_type, blowout_location, source_touchable],
    )


def test_retract_after_dispense_raises_for_invalid_retract_point(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
) -> None:
    """It should raise an error if the retract end point is deemed bad."""
    source_location = Location(Point(1, 2, 3), labware=None)
    source_well = decoy.mock(cls=WellCore)
    dest_well = decoy.mock(cls=WellCore)
    well_top_point = Point(1, 2, 3)
    well_bottom_point = Point(4, 5, 6)
    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=Location(Point(1, 1, 1), labware=None),
        target_well=dest_well,
        tip_state=TipState(),
        transfer_type=TransferType.ONE_TO_ONE,
    )
    decoy.when(dest_well.get_bottom(0)).then_return(well_bottom_point)
    decoy.when(dest_well.get_top(0)).then_return(well_top_point)
    decoy.when(source_well.get_top(0)).then_return(Point(10, 20, 30))
    decoy.when(
        tx_utils.raise_if_location_inside_liquid(
            location=Location(Point(12, 24, 36), labware=None),
            well_core=dest_well,
            location_check_descriptors=LocationCheckDescriptors(
                location_type="retract end",
                pipetting_action="dispense",
            ),
            logger=matchers.Anything(),
        )
    ).then_raise(RuntimeError("Oh no!"))
    with pytest.raises(RuntimeError, match="Oh no!"):
        subject.retract_after_dispensing(
            trash_location=Location(Point(), labware=None),
            source_location=source_location,
            source_well=source_well,
            add_final_air_gap=True,
        )


def test_retract_after_dispense_with_blowout_in_src_moves_to_safe_loc_for_air_gap(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
) -> None:
    """It should retract after a dispense and move to safe location for air gap at dest well and source well."""
    source_location = Location(Point(1, 2, 3), labware=None)
    source_well = decoy.mock(cls=WellCore)
    dest_well = decoy.mock(cls=WellCore)
    well_top_point = Point(1, 2, 3)
    well_bottom_point = Point(4, 5, 6)
    air_gap_volume = 0.123
    air_gap_flow_rate_by_vol = 123
    air_gap_correction_by_vol = 0.321

    sample_transfer_props.dispense.retract.air_gap_by_volume.set_for_volume(
        0, air_gap_volume
    )
    sample_transfer_props.aspirate.flow_rate_by_volume.set_for_volume(
        air_gap_volume, air_gap_flow_rate_by_vol
    )
    sample_transfer_props.aspirate.correction_by_volume.set_for_volume(
        air_gap_volume, air_gap_correction_by_vol
    )

    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=Location(Point(1, 1, 1), labware=None),
        target_well=dest_well,
        tip_state=TipState(),
        transfer_type=TransferType.ONE_TO_ONE,
    )
    decoy.when(mock_instrument_core.get_current_volume()).then_return(0)
    decoy.when(dest_well.get_bottom(0)).then_return(well_bottom_point)
    decoy.when(dest_well.get_top(0)).then_return(well_top_point)
    decoy.when(source_well.get_top(0)).then_return(Point(10, 20, 30))
    # Assume air gap safe location is above retract location
    decoy.when(dest_well.get_top(AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP)).then_return(
        Point(1, 2, 37)
    )
    # Assume air gap safe location for air-gapping at src is above blowout position,
    # where blowout position is source well top (as set above)
    decoy.when(source_well.get_top(AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP)).then_return(
        Point(10, 20, 31)
    )
    subject.retract_after_dispensing(
        trash_location=Location(Point(), labware=None),
        source_location=source_location,
        source_well=source_well,
        add_final_air_gap=True,
    )
    decoy.verify(
        mock_instrument_core.move_to(
            location=Location(Point(12, 24, 36), labware=None),
            well_core=dest_well,
            force_direct=True,
            minimum_z_height=None,
            speed=50,
        ),
        mock_instrument_core.delay(10),
        mock_instrument_core.touch_tip(
            location=Location(Point(12, 24, 36), labware=None),
            well_core=dest_well,
            radius=1,
            mm_from_edge=0.75,
            z_offset=-1,
            speed=30,
        ),
        mock_instrument_core.move_to(
            location=Location(Point(12, 24, 36), labware=None),
            well_core=dest_well,
            force_direct=True,
            minimum_z_height=None,
            speed=None,
        ),
        mock_instrument_core.move_to(
            location=Location(Point(12, 24, 37), labware=None),
            well_core=dest_well,
            force_direct=True,
            minimum_z_height=None,
            speed=None,
        ),
        mock_instrument_core.air_gap_in_place(
            volume=air_gap_volume,
            flow_rate=air_gap_flow_rate_by_vol,
            correction_volume=air_gap_correction_by_vol,
        ),
        mock_instrument_core.delay(0.2),
        mock_instrument_core.blow_out(
            location=Location(Point(10, 20, 30), labware=None),
            well_core=source_well,
            in_place=False,
            flow_rate=100,
        ),
        mock_instrument_core.touch_tip(
            location=Location(Point(10, 20, 30), labware=None),
            well_core=source_well,
            radius=1,
            mm_from_edge=0.75,
            z_offset=-1,
            speed=30,
        ),
        mock_instrument_core.move_to(
            location=Location(Point(10, 20, 30), labware=None),
            well_core=source_well,
            force_direct=True,
            minimum_z_height=None,
            speed=None,
        ),
        mock_instrument_core.move_to(
            location=Location(Point(10, 20, 31), labware=None),
            well_core=source_well,
            force_direct=True,
            minimum_z_height=None,
            speed=None,
        ),
        mock_instrument_core.prepare_to_aspirate(),
        mock_instrument_core.air_gap_in_place(
            # type: ignore[func-returns-value]
            volume=air_gap_volume,
            flow_rate=air_gap_flow_rate_by_vol,
            correction_volume=air_gap_correction_by_vol,
        ),
        mock_instrument_core.delay(0.2),  # type: ignore[func-returns-value]
    )


@pytest.mark.parametrize(
    argnames=["is_last_retract", "add_final_air_gap", "expect_air_gap"],
    argvalues=[
        (True, False, False),
        (False, False, False),
        (True, True, True),
        (False, True, True),
    ],
)
def test_multi_dispense_retract_after_dispense_without_conditioning_volume_or_blowout(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
    is_last_retract: bool,
    add_final_air_gap: bool,
    expect_air_gap: bool,
) -> None:
    """It should execute retract steps.

    Should execute steps, including expected air gaps, to retract from well during
    a multi-dispense when there's no conditioning volume, no disposal volume and no blowout.
    """
    source_location = Location(Point(1, 2, 3), labware=None)
    source_well = decoy.mock(cls=WellCore)
    dest_well = decoy.mock(cls=WellCore)
    well_top_point = Point(1, 2, 3)
    sample_transfer_props.multi_dispense.retract.touch_tip.enabled = True  # type: ignore[union-attr]
    air_gap_flow_rate_by_vol = 123
    air_gap_correction_by_vol = 0.321

    sample_transfer_props.multi_dispense.retract.air_gap_by_volume.set_for_volume(  # type: ignore[union-attr]
        100, 51
    )
    sample_transfer_props.aspirate.flow_rate_by_volume.set_for_volume(
        51, air_gap_flow_rate_by_vol
    )
    sample_transfer_props.aspirate.correction_by_volume.set_for_volume(
        51, air_gap_correction_by_vol
    )
    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=Location(Point(1, 1, 1), labware=None),
        target_well=dest_well,
        tip_state=TipState(
            ready_to_aspirate=True,
            last_liquid_and_air_gap_in_tip=LiquidAndAirGapPair(
                # Since we'll be testing retract_during_multi_dispensing right after
                # initializing the executor, this is the tip state that retract_during_multi_dispensing
                # will start with.
                liquid=100,
                air_gap=0,
            ),
        ),
        transfer_type=TransferType.ONE_TO_MANY,
    )
    decoy.when(mock_instrument_core.get_current_volume()).then_return(0)
    decoy.when(dest_well.get_top(0)).then_return(well_top_point)
    # Assume air gap safe location is below retract location
    decoy.when(dest_well.get_top(AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP)).then_return(
        Point(1, 2, 3)
    )
    subject.retract_during_multi_dispensing(
        trash_location=Location(Point(), labware=None),
        source_location=source_location,
        source_well=source_well,
        conditioning_volume=0,
        add_final_air_gap=add_final_air_gap,
        is_last_retract=is_last_retract,
    )
    decoy.verify(
        tx_utils.raise_if_location_inside_liquid(
            location=Location(Point(3, 5, 4), labware=None),
            well_core=dest_well,
            location_check_descriptors=LocationCheckDescriptors(
                location_type="retract end",
                pipetting_action="dispense",
            ),
            logger=matchers.Anything(),
        ),
        mock_instrument_core.move_to(
            location=Location(Point(3, 5, 4), labware=None),
            well_core=dest_well,
            force_direct=True,
            minimum_z_height=None,
            speed=50,
        ),
        mock_instrument_core.touch_tip(
            location=Location(Point(3, 5, 4), labware=None),
            well_core=dest_well,
            radius=1,
            mm_from_edge=0.5,
            z_offset=-1,
            speed=30,
        ),
        mock_instrument_core.move_to(
            location=Location(Point(3, 5, 4), labware=None),
            well_core=dest_well,
            force_direct=True,
            minimum_z_height=None,
            speed=None,
        ),
        *(
            expect_air_gap
            and [
                mock_instrument_core.air_gap_in_place(
                    # type: ignore[func-returns-value]
                    volume=51,
                    flow_rate=air_gap_flow_rate_by_vol,
                    correction_volume=air_gap_correction_by_vol,
                ),
                mock_instrument_core.delay(0.2),  # type: ignore[func-returns-value]
            ]
            or []
        ),
    )


@pytest.mark.parametrize(
    argnames=[
        "is_last_retract",
        "add_final_air_gap",
        "expect_blowout",
        "expect_air_gap",
    ],
    argvalues=[
        (True, False, True, False),
        (False, False, False, False),
        (True, True, True, True),
        (False, True, False, True),
    ],
)
def test_multi_dispense_retract_after_dispense_with_blowout_without_conditioning_volume(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
    is_last_retract: bool,
    add_final_air_gap: bool,
    expect_air_gap: bool,
    expect_blowout: bool,
) -> None:
    """It should execute retract steps.

    Should execute steps, including expected air gaps, to retract from well during a
    multi-dispense when there's no conditioning volume while blowout is enabled.
    """
    source_location = Location(Point(1, 2, 3), labware=None)
    source_well = decoy.mock(cls=WellCore)
    dest_well = decoy.mock(cls=WellCore)
    well_top_point = Point(1, 2, 3)
    air_gap_volume = 0.123
    air_gap_flow_rate_by_vol = 123
    air_gap_correction_by_vol = 0.321

    sample_transfer_props.multi_dispense.retract.air_gap_by_volume.set_for_all_volumes(  # type: ignore[union-attr]
        air_gap_volume
    )
    sample_transfer_props.aspirate.flow_rate_by_volume.set_for_volume(
        air_gap_volume, air_gap_flow_rate_by_vol
    )
    sample_transfer_props.aspirate.correction_by_volume.set_for_volume(
        air_gap_volume, air_gap_correction_by_vol
    )

    sample_transfer_props.multi_dispense.retract.touch_tip.enabled = True  # type: ignore[union-attr]
    sample_transfer_props.multi_dispense.retract.blowout.enabled = True  # type: ignore[union-attr]

    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=Location(Point(1, 1, 1), labware=None),
        target_well=dest_well,
        tip_state=TipState(),
        transfer_type=TransferType.ONE_TO_MANY,
    )
    decoy.when(mock_instrument_core.get_current_volume()).then_return(0)
    decoy.when(dest_well.get_top(0)).then_return(well_top_point)
    # Assume air gap safe location is below retract location
    decoy.when(dest_well.get_top(AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP)).then_return(
        Point(1, 2, 3)
    )
    subject.retract_during_multi_dispensing(
        trash_location=Location(Point(), labware=None),
        source_location=source_location,
        source_well=source_well,
        conditioning_volume=0,
        add_final_air_gap=add_final_air_gap,
        is_last_retract=is_last_retract,
    )
    decoy.verify(
        tx_utils.raise_if_location_inside_liquid(
            location=Location(Point(3, 5, 4), labware=None),
            well_core=dest_well,
            location_check_descriptors=LocationCheckDescriptors(
                location_type="retract end",
                pipetting_action="dispense",
            ),
            logger=matchers.Anything(),
        ),
        mock_instrument_core.move_to(
            location=Location(Point(3, 5, 4), labware=None),
            well_core=dest_well,
            force_direct=True,
            minimum_z_height=None,
            speed=50,
        ),
        *(
            expect_blowout
            and [
                mock_instrument_core.blow_out(  # type: ignore[func-returns-value]
                    location=Location(Point(3, 5, 4), labware=None),
                    well_core=dest_well,
                    in_place=True,
                    flow_rate=10,
                )
            ]
            or []
        ),
        mock_instrument_core.touch_tip(
            location=Location(Point(3, 5, 4), labware=None),
            well_core=dest_well,
            radius=1,
            mm_from_edge=0.5,
            z_offset=-1,
            speed=30,
        ),
        mock_instrument_core.move_to(
            location=Location(Point(3, 5, 4), labware=None),
            well_core=dest_well,
            force_direct=True,
            minimum_z_height=None,
            speed=None,
        ),
        *(
            expect_air_gap
            and [
                mock_instrument_core.air_gap_in_place(
                    # type: ignore[func-returns-value]
                    volume=air_gap_volume,
                    flow_rate=air_gap_flow_rate_by_vol,
                    correction_volume=air_gap_correction_by_vol,
                ),
                mock_instrument_core.delay(0.2),  # type: ignore[func-returns-value]
            ]
            or []
        ),
    )
    assert subject.tip_state.last_liquid_and_air_gap_in_tip == LiquidAndAirGapPair(
        air_gap=air_gap_volume if expect_air_gap else 0, liquid=0
    )


def test_multi_dispense_retract_raises_for_invalid_retract_point(
    decoy: Decoy,
    mock_instrument_core: InstrumentCore,
    sample_transfer_props: TransferProperties,
) -> None:
    """It should raise an error if the retract end point is deemed bad."""
    source_location = Location(Point(1, 2, 3), labware=None)
    source_well = decoy.mock(cls=WellCore)
    dest_well = decoy.mock(cls=WellCore)
    well_top_point = Point(1, 2, 3)
    subject = TransferComponentsExecutor(
        instrument_core=mock_instrument_core,
        transfer_properties=sample_transfer_props,
        target_location=Location(Point(1, 1, 1), labware=None),
        target_well=dest_well,
        tip_state=TipState(),
        transfer_type=TransferType.ONE_TO_MANY,
    )
    decoy.when(dest_well.get_top(0)).then_return(well_top_point)
    decoy.when(
        tx_utils.raise_if_location_inside_liquid(
            location=Location(Point(3, 5, 4), labware=None),
            well_core=dest_well,
            location_check_descriptors=LocationCheckDescriptors(
                location_type="retract end",
                pipetting_action="dispense",
            ),
            logger=matchers.Anything(),
        )
    ).then_raise(RuntimeError("Oh no!"))
    with pytest.raises(RuntimeError, match="Oh no!"):
        subject.retract_during_multi_dispensing(
            trash_location=Location(Point(), labware=None),
            source_location=source_location,
            source_well=source_well,
            conditioning_volume=0,
            add_final_air_gap=True,
            is_last_retract=False,
        )


@pytest.mark.parametrize(
    argnames=["position_reference", "offset", "expected_result"],
    argvalues=[
        (PositionReference.WELL_TOP, Coordinate(x=11, y=12, z=13), Point(12, 14, 16)),
        (
            PositionReference.WELL_BOTTOM,
            Coordinate(x=21, y=22, z=23),
            Point(25, 27, 29),
        ),
        (
            PositionReference.WELL_CENTER,
            Coordinate(x=31, y=32, z=33),
            Point(38, 40, 42),
        ),
        (
            PositionReference.LIQUID_MENISCUS,
            Coordinate(x=41, y=42, z=43),
            Point(45, 47, 61),
        ),
    ],
)
def test_absolute_point_from_position_reference_and_offset(
    decoy: Decoy,
    position_reference: PositionReference,
    offset: Coordinate,
    expected_result: Point,
) -> None:
    """It should return the correct absolute point based on well, position reference and offset."""
    well = decoy.mock(cls=WellCore)

    well_top_point = Point(1, 2, 3)
    well_bottom_point = Point(4, 5, 6)
    well_center_point = Point(7, 8, 9)
    estimated_liquid_height = 12
    decoy.when(well.get_bottom(0)).then_return(well_bottom_point)
    decoy.when(well.get_top(0)).then_return(well_top_point)
    decoy.when(well.get_center()).then_return(well_center_point)
    decoy.when(
        well.estimate_liquid_height_after_pipetting(
            operation_volume=123, mount=Mount.RIGHT
        ),
    ).then_return(estimated_liquid_height)
    decoy.when(well.get_bottom(12)).then_return(Point(4, 5, 18))

    assert (
        absolute_point_from_position_reference_and_offset(
            well=well,
            well_volume_difference=123,
            position_reference=position_reference,
            offset=offset,
            mount=Mount.RIGHT,
        )
        == expected_result
    )


def test_absolute_point_from_position_reference_and_offset_raises_errors(
    decoy: Decoy,
) -> None:
    """It should raise errors for invalid input."""
    well = decoy.mock(cls=WellCore)
    with pytest.raises(ValueError, match="Unknown position reference"):
        absolute_point_from_position_reference_and_offset(
            well=well,
            well_volume_difference=123,
            position_reference="PositionReference",  # type: ignore[arg-type]
            offset=Coordinate(x=0, y=0, z=0),
            mount=Mount.RIGHT,
        )



from opentrons_shared_data.labware.types import LabwareDefinition2

from opentrons.protocol_api import Labware
from opentrons.protocol_api.core.legacy.legacy_labware_core import LegacyLabwareCore
from opentrons.types import Point


def test_wells_accessor(
    min_lw: Labware,
    min_lw_impl: LegacyLabwareCore,
    minimal_labware_def: LabwareDefinition2,
) -> None:
    depth1 = minimal_labware_def["wells"]["A1"]["depth"]
    depth2 = minimal_labware_def["wells"]["A2"]["depth"]
    x = minimal_labware_def["wells"]["A2"]["x"]
    y = minimal_labware_def["wells"]["A2"]["y"]
    offset = min_lw_impl.get_geometry()._offset
    a1 = Point(x=offset[0], y=offset[1], z=offset[2] + depth1)
    a2 = Point(x=offset[0] + x, y=offset[1] + y, z=offset[2] + depth2)
    assert min_lw.wells()[0].geometry._position == a1
    assert min_lw.wells()[1].geometry._position == a2


def test_wells_name_accessor(
    min_lw: Labware,
    min_lw_impl: LegacyLabwareCore,
    minimal_labware_def: LabwareDefinition2,
) -> None:
    depth1 = minimal_labware_def["wells"]["A1"]["depth"]
    depth2 = minimal_labware_def["wells"]["A2"]["depth"]
    x = minimal_labware_def["wells"]["A2"]["x"]
    y = minimal_labware_def["wells"]["A2"]["y"]
    offset = min_lw_impl.get_geometry().offset
    a1 = Point(x=offset[0], y=offset[1], z=offset[2] + depth1)
    a2 = Point(x=offset[0] + x, y=offset[1] + y, z=offset[2] + depth2)
    assert min_lw.wells_by_name()["A1"].geometry._position == a1
    assert min_lw.wells_by_name()["A2"].geometry._position == a2


def test_deprecated_index_accessors(min_lw: Labware) -> None:
    assert min_lw.wells_by_name() == min_lw.wells_by_index()
    assert min_lw.rows_by_name() == min_lw.rows_by_index()
    assert min_lw.columns_by_name() == min_lw.columns_by_index()


def test_dict_accessor(
    min_lw: Labware,
    min_lw_impl: LegacyLabwareCore,
    minimal_labware_def: LabwareDefinition2,
) -> None:
    depth1 = minimal_labware_def["wells"]["A1"]["depth"]
    depth2 = minimal_labware_def["wells"]["A2"]["depth"]
    x = minimal_labware_def["wells"]["A2"]["x"]
    y = minimal_labware_def["wells"]["A2"]["y"]
    offset = min_lw_impl.get_geometry().offset
    a1 = Point(x=offset[0], y=offset[1], z=offset[2] + depth1)
    a2 = Point(x=offset[0] + x, y=offset[1] + y, z=offset[2] + depth2)
    assert min_lw["A1"].geometry._position == a1
    assert min_lw["A2"].geometry._position == a2


def test_rows_accessor(
    min_lw2_impl: LegacyLabwareCore,
    min_lw2: Labware,
    minimal_labware_def2: LabwareDefinition2,
) -> None:
    depth1 = minimal_labware_def2["wells"]["A1"]["depth"]
    x1 = minimal_labware_def2["wells"]["A1"]["x"]
    y1 = minimal_labware_def2["wells"]["A1"]["y"]
    depth2 = minimal_labware_def2["wells"]["B2"]["depth"]
    x2 = minimal_labware_def2["wells"]["B2"]["x"]
    y2 = minimal_labware_def2["wells"]["B2"]["y"]
    offset = min_lw2_impl.get_geometry().offset
    a1 = Point(x=offset[0] + x1, y=offset[1] + y1, z=offset[2] + depth1)
    b2 = Point(x=offset[0] + x2, y=offset[1] + y2, z=offset[2] + depth2)
    assert min_lw2.rows()[0][0].geometry._position == a1
    assert min_lw2.rows()[1][1].geometry._position == b2


def test_row_name_accessor(
    min_lw2_impl: LegacyLabwareCore,
    min_lw2: Labware,
    minimal_labware_def2: LabwareDefinition2,
) -> None:
    depth1 = minimal_labware_def2["wells"]["A1"]["depth"]
    x1 = minimal_labware_def2["wells"]["A1"]["x"]
    y1 = minimal_labware_def2["wells"]["A1"]["y"]
    depth2 = minimal_labware_def2["wells"]["B2"]["depth"]
    x2 = minimal_labware_def2["wells"]["B2"]["x"]
    y2 = minimal_labware_def2["wells"]["B2"]["y"]
    offset = min_lw2_impl.get_geometry().offset
    a1 = Point(x=offset[0] + x1, y=offset[1] + y1, z=offset[2] + depth1)
    b2 = Point(x=offset[0] + x2, y=offset[1] + y2, z=offset[2] + depth2)
    assert min_lw2.rows_by_name()["A"][0].geometry._position == a1
    assert min_lw2.rows_by_name()["B"][1].geometry._position == b2


def test_cols_accessor(
    min_lw_impl: LegacyLabwareCore,
    min_lw: Labware,
    minimal_labware_def: LabwareDefinition2,
) -> None:
    depth1 = minimal_labware_def["wells"]["A1"]["depth"]
    depth2 = minimal_labware_def["wells"]["A2"]["depth"]
    x = minimal_labware_def["wells"]["A2"]["x"]
    y = minimal_labware_def["wells"]["A2"]["y"]
    offset = min_lw_impl.get_geometry().offset
    a1 = Point(x=offset[0], y=offset[1], z=offset[2] + depth1)
    a2 = Point(x=offset[0] + x, y=offset[1] + y, z=offset[2] + depth2)
    assert min_lw.columns()[0][0].geometry._position == a1
    assert min_lw.columns()[1][0].geometry._position == a2


def test_col_name_accessor(
    min_lw: Labware,
    min_lw_impl: LegacyLabwareCore,
    minimal_labware_def: LabwareDefinition2,
) -> None:
    depth1 = minimal_labware_def["wells"]["A1"]["depth"]
    depth2 = minimal_labware_def["wells"]["A2"]["depth"]
    x = minimal_labware_def["wells"]["A2"]["x"]
    y = minimal_labware_def["wells"]["A2"]["y"]
    offset = min_lw_impl.get_geometry().offset
    a1 = Point(x=offset[0], y=offset[1], z=offset[2] + depth1)
    a2 = Point(x=offset[0] + x, y=offset[1] + y, z=offset[2] + depth2)
    assert min_lw.columns_by_name()["1"][0].geometry._position == a1
    assert min_lw.columns_by_name()["2"][0].geometry._position == a2



"""Test absorbance reader initilize command."""

import math
from typing import Dict, List, Optional

import pytest
from decoy import Decoy

from opentrons.drivers.types import ABSMeasurementConfig, ABSMeasurementMode
from opentrons.hardware_control.modules import AbsorbanceReader
from opentrons.protocol_engine.commands.absorbance_reader import (
    ReadAbsorbanceParams,
    ReadAbsorbanceResult,
)
from opentrons.protocol_engine.commands.absorbance_reader.read import (
    ReadAbsorbanceImpl,
)
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.errors import (
    CannotPerformModuleAction,
)
from opentrons.protocol_engine.execution import EquipmentHandler
from opentrons.protocol_engine.resources import FileProvider
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.module_substates import (
    AbsorbanceReaderId,
    AbsorbanceReaderSubState,
)
from opentrons.protocol_engine.state.modules import ModuleView
from opentrons.protocol_engine.state.state import StateView


def _get_absorbance_map(data: Optional[List[float]] = None) -> Dict[str, float]:
    raw_values = (data or [0] * 96).copy()
    raw_values.reverse()
    well_map: Dict[str, float] = {}
    for i, value in enumerate(raw_values):
        row = chr(ord("A") + i // 12)  # Convert index to row (A-H)
        col = (i % 12) + 1  # Convert index to column (1-12)
        well_key = f"{row}{col}"
        # Truncate the value to the third decimal place
        well_map[well_key] = max(math.floor(value * 1000) / 1000, 0)
    return well_map


async def test_absorbance_reader_implementation(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
    file_provider: FileProvider,
) -> None:
    """It should validate, find hardware module if not virtualized, and disengage."""
    subject = ReadAbsorbanceImpl(
        state_view=state_view, equipment=equipment, file_provider=file_provider
    )

    params = ReadAbsorbanceParams(
        moduleId="unverified-module-id",
    )

    mabsorbance_module_substate = decoy.mock(cls=AbsorbanceReaderSubState)
    absorbance_module_hw = decoy.mock(cls=AbsorbanceReader)
    verified_module_id = AbsorbanceReaderId("module-id")
    asbsorbance_result = {1: {"A1": 1.2}}

    decoy.when(
        state_view.modules.get_absorbance_reader_substate("unverified-module-id")
    ).then_return(mabsorbance_module_substate)

    decoy.when(mabsorbance_module_substate.module_id).then_return(verified_module_id)

    decoy.when(equipment.get_module_hardware_api(verified_module_id)).then_return(
        absorbance_module_hw
    )

    decoy.when(await absorbance_module_hw.start_measure()).then_return([[1.2, 1.3]])
    decoy.when(absorbance_module_hw._measurement_config).then_return(
        ABSMeasurementConfig(
            measure_mode=ABSMeasurementMode.SINGLE,
            sample_wavelengths=[1, 2],
            reference_wavelength=None,
        )
    )
    decoy.when(
        state_view.modules.convert_absorbance_reader_data_points([1.2, 1.3])
    ).then_return({"A1": 1.2})

    result = await subject.execute(params=params)

    assert result == SuccessData(
        public=ReadAbsorbanceResult(
            data=asbsorbance_result,
            fileIds=[],
        ),
        state_update=update_types.StateUpdate(
            files_added=update_types.FilesAddedUpdate(file_ids=[]),
            absorbance_reader_state_update=update_types.AbsorbanceReaderStateUpdate(
                module_id="module-id",
                absorbance_reader_data=update_types.AbsorbanceReaderDataUpdate(
                    read_result=asbsorbance_result
                ),
            ),
        ),
    )


async def test_convert_absorbance_reader_data_points() -> None:
    """It should validate and convert the absorbance reader values."""
    # Test valid values
    raw_data = (
        [0.04877041280269623, 0.046341221779584885]
        + [0.43] * 92  # fill rest of the values with 0.43
        + [0.03789025545120239, 2.8744750022888184]
    )
    expected = _get_absorbance_map(raw_data)
    converted = ModuleView.convert_absorbance_reader_data_points(raw_data)
    assert len(converted) == 96
    assert converted == expected
    assert converted["A1"] == 2.874
    assert converted["A2"] == 0.037
    assert converted["E1"] == 0.43
    assert converted["H12"] == 0.048  # the data is flipped, so arr[0] == H12

    # Test near-zero values in scientic notation
    raw_data = (
        [0.24877041280269623, -9.5000e-9]
        + [0.11] * 92  # fill rest of the values with 0.11
        + [1.3489025545120239, 8.2987e-9]
    )
    expected = _get_absorbance_map(raw_data)
    converted = ModuleView.convert_absorbance_reader_data_points(raw_data)
    assert len(converted) == 96
    assert converted == expected
    assert converted["A1"] == 0.0
    assert converted["A2"] == 1.348
    assert converted["E1"] == 0.11
    assert converted["H11"] == 0.0  # tests the clamp-to-0 behaviora
    assert converted["H12"] == 0.248  # the data is flipped, so arr[0] == H12

    # Test invalid data len 1
    with pytest.raises(ValueError):
        ModuleView.convert_absorbance_reader_data_points([0])

    # Test invalid data len 107
    with pytest.raises(ValueError):
        ModuleView.convert_absorbance_reader_data_points([1] * 107)


async def test_read_raises_cannot_preform_action(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
    file_provider: FileProvider,
) -> None:
    """It should raise CannotPerformModuleAction when not configured/lid is not on."""
    subject = ReadAbsorbanceImpl(
        state_view=state_view, equipment=equipment, file_provider=file_provider
    )

    params = ReadAbsorbanceParams(
        moduleId="unverified-module-id",
    )

    mabsorbance_module_substate = decoy.mock(cls=AbsorbanceReaderSubState)
    absorbance_module_hw = decoy.mock(cls=AbsorbanceReader)

    verified_module_id = AbsorbanceReaderId("module-id")

    decoy.when(
        state_view.modules.get_absorbance_reader_substate("unverified-module-id")
    ).then_return(mabsorbance_module_substate)

    decoy.when(mabsorbance_module_substate.module_id).then_return(verified_module_id)

    decoy.when(equipment.get_module_hardware_api(verified_module_id)).then_return(
        absorbance_module_hw
    )

    decoy.when(mabsorbance_module_substate.configured).then_return(False)

    with pytest.raises(CannotPerformModuleAction):
        await subject.execute(params=params)

    decoy.when(mabsorbance_module_substate.configured).then_return(True)

    decoy.when(mabsorbance_module_substate.is_lid_on).then_return(False)

    with pytest.raises(CannotPerformModuleAction):
        await subject.execute(params=params)



"""Hardware Modules' substates."""

from typing import Union

from .absorbance_reader_substate import AbsorbanceReaderId, AbsorbanceReaderSubState
from .flex_stacker_substate import FlexStackerId, FlexStackerSubState
from .heater_shaker_module_substate import (
    HeaterShakerModuleId,
    HeaterShakerModuleSubState,
)
from .magnetic_block_substate import MagneticBlockId, MagneticBlockSubState
from .magnetic_module_substate import MagneticModuleId, MagneticModuleSubState
from .temperature_module_substate import TemperatureModuleId, TemperatureModuleSubState
from .thermocycler_module_substate import (
    ThermocyclerModuleId,
    ThermocyclerModuleSubState,
)

ModuleSubStateType = Union[
    HeaterShakerModuleSubState,
    MagneticModuleSubState,
    TemperatureModuleSubState,
    ThermocyclerModuleSubState,
    MagneticBlockSubState,
    AbsorbanceReaderSubState,
    FlexStackerSubState,
]

__all__ = [
    "MagneticModuleSubState",
    "MagneticModuleId",
    "HeaterShakerModuleSubState",
    "HeaterShakerModuleId",
    "TemperatureModuleSubState",
    "TemperatureModuleId",
    "ThermocyclerModuleSubState",
    "ThermocyclerModuleId",
    "MagneticBlockSubState",
    "MagneticBlockId",
    "AbsorbanceReaderSubState",
    "AbsorbanceReaderId",
    "FlexStackerSubState",
    "FlexStackerId",
    # Union of all module substates
    "ModuleSubStateType",
]



import pytest
from mock import AsyncMock

from opentrons.drivers.asyncio.communication.errors import UnhandledGcode
from opentrons.drivers.asyncio.communication.serial_connection import (
    SerialConnection,
)
from opentrons.drivers.command_builder import CommandBuilder
from opentrons.drivers.temp_deck.driver import (
    TEMP_DECK_COMMAND_TERMINATOR,
    TempDeckDriver,
)
from opentrons.drivers.types import Temperature
from opentrons.drivers.utils import TEMPDECK_GCODE_ROUNDING_PRECISION


@pytest.fixture
def connection() -> AsyncMock:
    return AsyncMock(spec=SerialConnection)


@pytest.fixture
def driver(connection: AsyncMock) -> TempDeckDriver:
    connection.send_command.return_value = ""
    return TempDeckDriver(connection)


async def test_deactivate(driver: TempDeckDriver, connection: AsyncMock) -> None:
    """It should send a deactivate command"""
    await driver.deactivate()

    expected = CommandBuilder(terminator=TEMP_DECK_COMMAND_TERMINATOR).add_gcode("M18")

    connection.send_command.assert_called_once_with(command=expected, retries=3)


async def test_set_temperature(driver: TempDeckDriver, connection: AsyncMock) -> None:
    """It should send a set temperature command"""
    await driver.set_temperature(celsius=123.4444)

    expected = (
        CommandBuilder(terminator=TEMP_DECK_COMMAND_TERMINATOR)
        .add_gcode("M104")
        .add_float(
            prefix="S", value=123.4444, precision=TEMPDECK_GCODE_ROUNDING_PRECISION
        )
    )

    connection.send_command.assert_called_once_with(command=expected, retries=3)


async def test_get_temperature(driver: TempDeckDriver, connection: AsyncMock) -> None:
    """It should send a get temperature command and parse response"""
    connection.send_command.return_value = "T:132 C:25 ok\r\nok\r\n"

    response = await driver.get_temperature()

    expected = CommandBuilder(terminator=TEMP_DECK_COMMAND_TERMINATOR).add_gcode("M105")

    connection.send_command.assert_called_once_with(command=expected, retries=3)

    assert response == Temperature(current=25, target=132)


async def test_get_device_info_with_reset_reason(
    driver: TempDeckDriver, connection: AsyncMock
) -> None:
    """It should send a get device info command and parse response"""
    connection.send_command.return_value = "serial:s model:m version:v"

    response = await driver.get_device_info()

    device_info = CommandBuilder(terminator=TEMP_DECK_COMMAND_TERMINATOR).add_gcode(
        "M115"
    )
    reset_reason = CommandBuilder(terminator=TEMP_DECK_COMMAND_TERMINATOR).add_gcode(
        "M114"
    )

    connection.send_command.assert_any_call(command=device_info, retries=3)
    connection.send_command.assert_called_with(command=reset_reason, retries=3)

    assert response == {"serial": "s", "model": "m", "version": "v"}


async def test_get_device_info_no_reset_reason(
    driver: TempDeckDriver, connection: AsyncMock
) -> None:
    """It should send a get device info command and parse response"""

    async def fake_send_command(
        command: CommandBuilder, retries: int = 0, timeout: float | None = None
    ) -> str:
        if command.build().startswith("M114"):
            raise UnhandledGcode(
                port="fake port",
                response="ERR003: Unhandled Gcode",
                command=command.build(),
            )
        else:
            return "serial:s model:m version:v"

    connection.send_command.side_effect = fake_send_command

    response = await driver.get_device_info()

    device_info = CommandBuilder(terminator=TEMP_DECK_COMMAND_TERMINATOR).add_gcode(
        "M115"
    )
    reset_reason = CommandBuilder(terminator=TEMP_DECK_COMMAND_TERMINATOR).add_gcode(
        "M114"
    )

    connection.send_command.assert_any_call(command=device_info, retries=3)
    connection.send_command.assert_called_with(command=reset_reason, retries=3)

    assert response == {"serial": "s", "model": "m", "version": "v"}


async def test_enter_programming_mode(
    driver: TempDeckDriver, connection: AsyncMock
) -> None:
    """It should send an enter programming mode command"""
    await driver.enter_programming_mode()

    expected = CommandBuilder(terminator=TEMP_DECK_COMMAND_TERMINATOR).add_gcode("dfu")

    connection.send_command.assert_called_once_with(command=expected, retries=3)



"""Simple functional tests for the ModelUtils provider."""

import re

from opentrons.protocol_engine.resources import ModelUtils

RE_UUID = re.compile(
    r"^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$",
    flags=re.IGNORECASE,
)


def test_model_utils_generates_uuid() -> None:
    """It should generate a string matching a UUID."""
    result = ModelUtils().generate_id()

    assert RE_UUID.match(result)


def test_model_utils_generates_unique_id() -> None:
    """Generated IDs should be unique."""
    results = [ModelUtils().generate_id() for i in range(1000)]
    unique_results = set(results)

    assert len(results) == len(unique_results)


def test_model_utils_generates_id_with_prefix() -> None:
    """It should return an ID with the given prefix."""
    assert ModelUtils().generate_id(prefix="my-prefix-").startswith("my-prefix-")


def test_model_utils_ensures_id() -> None:
    """It should generate unique IDs only when needed."""
    subject = ModelUtils()
    assert subject.ensure_id("hello world") == "hello world"
    assert RE_UUID.match(subject.ensure_id())
    assert RE_UUID.match(subject.ensure_id(None))



from typing import AsyncIterator, Iterator

import pytest

from .build_module import build_module
from opentrons.hardware_control import ExecutionManager
from opentrons.hardware_control.emulation.settings import Settings
from opentrons.hardware_control.modules import MagDeck


@pytest.fixture
async def magdeck(
    emulation_app: Iterator[None],
    emulator_settings: Settings,
    execution_manager: ExecutionManager,
) -> AsyncIterator[MagDeck]:
    module = await build_module(
        MagDeck,
        port=emulator_settings.magdeck_proxy.driver_port,
        execution_manager=execution_manager,
    )
    try:
        yield module
    finally:
        await module.cleanup()


def test_device_info(magdeck: MagDeck) -> None:
    assert magdeck.device_info == {
        "model": "mag_deck_v20",
        "serial": "magnetic_emulator",
        "version": "2.0.0",
    }


async def test_engage_cycle(magdeck: MagDeck) -> None:
    """It should cycle engage and disengage"""
    await magdeck.engage(1)
    assert magdeck.current_height == 1
    assert magdeck.live_data == {
        "data": {"engaged": True, "height": 1.0},
        "status": "engaged",
    }

    await magdeck.deactivate()
    assert magdeck.current_height == 0
    assert magdeck.live_data == {
        "data": {"engaged": False, "height": 0.0},
        "status": "disengaged",
    }


async def test_engage_from_base_cycle(magdeck: MagDeck) -> None:
    """It should cycle engage/disengage, taking the offset from base into account."""
    await magdeck.engage(height_from_base=1)
    assert magdeck.current_height == 3.5
    assert magdeck.live_data == {
        "data": {"engaged": True, "height": 3.5},
        "status": "engaged",
    }

    await magdeck.deactivate()
    assert magdeck.current_height == 0
    assert magdeck.live_data == {
        "data": {"engaged": False, "height": 0.0},
        "status": "disengaged",
    }


async def test_forcible_deactivate(
    magdeck: MagDeck, execution_manager: ExecutionManager
) -> None:
    """Can override wait_for_is_running."""
    await execution_manager.pause()
    await magdeck.deactivate(must_be_running=False)



"""Tests for opentrons.legacy.Deck."""

import inspect
from typing import Dict, cast

import pytest
from decoy import Decoy

from opentrons_shared_data.deck.types import DeckDefinitionV5, SlotDefV3

from opentrons.motion_planning import adjacent_slots_getters as mock_adjacent_slots
from opentrons.protocol_api import (
    OFF_DECK,
    Deck,
    Labware,
)
from opentrons.protocol_api import (
    validation as mock_validation,
)
from opentrons.protocol_api.core.common import LabwareCore, ModuleCore, ProtocolCore
from opentrons.protocol_api.core.core_map import LoadedCoreMap
from opentrons.protocol_api.deck import CalibrationPosition
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.api_support.util import APIVersionError
from opentrons.types import DeckSlotName, Point


@pytest.fixture
def deck_definition() -> DeckDefinitionV5:
    """Get a deck definition value object."""
    return cast(
        DeckDefinitionV5,
        {
            "locations": {"addressableAreas": [], "calibrationPoints": []},
            "cutoutFixtures": {},
        },
    )


@pytest.fixture
def api_version() -> APIVersion:
    """Get a dummy `APIVersion` with which to configure the subject."""
    return APIVersion(1, 234)


@pytest.fixture(autouse=True)
def _mock_validation_module(decoy: Decoy, monkeypatch: pytest.MonkeyPatch) -> None:
    """Mock out opentrons.legacy.validation functions."""
    for name, func in inspect.getmembers(mock_validation, inspect.isfunction):
        monkeypatch.setattr(mock_validation, name, decoy.mock(func=func))


@pytest.fixture(autouse=True)
def _mock_adjacent_slots_module(decoy: Decoy, monkeypatch: pytest.MonkeyPatch) -> None:
    """Mock out opentrons.motion_planning.adjacent_slots_getters functions."""
    for name, func in inspect.getmembers(mock_adjacent_slots, inspect.isfunction):
        monkeypatch.setattr(mock_adjacent_slots, name, decoy.mock(func=func))


@pytest.fixture
def mock_protocol_core(decoy: Decoy) -> ProtocolCore:
    """Get a mock ProtocolCore."""
    return decoy.mock(cls=ProtocolCore)


@pytest.fixture
def mock_core_map(decoy: Decoy) -> LoadedCoreMap:
    """Get a mock LoadedCoreMap."""
    return decoy.mock(cls=LoadedCoreMap)


@pytest.fixture
def slot_definitions_by_name() -> Dict[str, SlotDefV3]:
    """Get a dictionary of slot names to slot definitions."""
    return {"1": {}}


@pytest.fixture
def staging_slot_definitions_by_name() -> Dict[str, SlotDefV3]:
    """Get a dictionary of staging slot names to slot definitions."""
    return {"2": {}}


@pytest.fixture
def subject(
    decoy: Decoy,
    deck_definition: DeckDefinitionV5,
    mock_protocol_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    api_version: APIVersion,
    slot_definitions_by_name: Dict[str, SlotDefV3],
    staging_slot_definitions_by_name: Dict[str, SlotDefV3],
) -> Deck:
    """Get a Deck test subject with its dependencies mocked out."""
    decoy.when(mock_protocol_core.get_deck_definition()).then_return(deck_definition)
    decoy.when(mock_protocol_core.get_slot_definitions()).then_return(
        slot_definitions_by_name
    )
    decoy.when(mock_protocol_core.get_staging_slot_definitions()).then_return(
        staging_slot_definitions_by_name
    )

    return Deck(
        protocol_core=mock_protocol_core,
        core_map=mock_core_map,
        api_version=api_version,
    )


def test_get_empty_slot(
    decoy: Decoy,
    mock_protocol_core: ProtocolCore,
    api_version: APIVersion,
    subject: Deck,
) -> None:
    """It should return None for slots if empty."""
    decoy.when(mock_protocol_core.robot_type).then_return("OT-3 Standard")
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(42, api_version, "OT-3 Standard")
    ).then_return(DeckSlotName.SLOT_2)
    decoy.when(mock_protocol_core.get_slot_item(DeckSlotName.SLOT_2)).then_return(None)

    assert subject[42] is None


def test_get_slot_invalid_key(
    decoy: Decoy,
    mock_protocol_core: ProtocolCore,
    api_version: APIVersion,
    subject: Deck,
) -> None:
    """It should map a ValueError from validation to a KeyError."""
    decoy.when(mock_protocol_core.robot_type).then_return("OT-3 Standard")
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(1, api_version, "OT-3 Standard")
    ).then_raise(TypeError("uh oh"))
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(2, api_version, "OT-3 Standard")
    ).then_raise(ValueError("oh no"))

    with pytest.raises(KeyError, match="1"):
        subject[1]

    with pytest.raises(KeyError, match="2"):
        subject[2]


def test_get_slot_item(
    decoy: Decoy,
    mock_protocol_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    api_version: APIVersion,
    subject: Deck,
) -> None:
    """It should map a ValueError from validation to a KeyError."""
    mock_labware_core = decoy.mock(cls=LabwareCore)
    mock_labware = decoy.mock(cls=Labware)

    decoy.when(mock_protocol_core.robot_type).then_return("OT-3 Standard")
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(42, api_version, "OT-3 Standard")
    ).then_return(DeckSlotName.SLOT_2)
    decoy.when(mock_protocol_core.get_slot_item(DeckSlotName.SLOT_2)).then_return(
        mock_labware_core
    )
    decoy.when(mock_core_map.get(mock_labware_core)).then_return(mock_labware)

    assert subject[42] is mock_labware


def test_delitem_aliases_to_move_labware(
    decoy: Decoy,
    mock_protocol_core: ProtocolCore,
    api_version: APIVersion,
    subject: Deck,
) -> None:
    """It should be equivalent to a manual labware move to off-deck, without pausing."""
    mock_labware_core = decoy.mock(cls=LabwareCore)

    decoy.when(mock_protocol_core.robot_type).then_return("OT-3 Standard")
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(42, api_version, "OT-3 Standard")
    ).then_return(DeckSlotName.SLOT_2)
    decoy.when(mock_protocol_core.get_slot_item(DeckSlotName.SLOT_2)).then_return(
        mock_labware_core
    )

    del subject[42]

    decoy.verify(
        mock_protocol_core.move_labware(
            mock_labware_core,
            OFF_DECK,
            use_gripper=False,
            pause_for_manual_move=False,
            pick_up_offset=None,
            drop_offset=None,
        )
    )


@pytest.mark.parametrize("api_version", [APIVersion(2, 14)])
def test_delitem_raises_on_api_2_14(
    subject: Deck,
) -> None:
    """It should raise on apiLevel 2.14."""
    with pytest.raises(APIVersionError):
        del subject[1]


def test_delitem_noops_if_slot_is_empty(
    decoy: Decoy,
    mock_protocol_core: ProtocolCore,
    api_version: APIVersion,
    subject: Deck,
) -> None:
    """It should do nothing, and not raise anything, if you try to delete from an empty slot."""
    decoy.when(mock_protocol_core.robot_type).then_return("OT-3 Standard")
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(1, api_version, "OT-3 Standard")
    ).then_return(DeckSlotName.SLOT_1)
    decoy.when(mock_protocol_core.get_slot_item(DeckSlotName.SLOT_1)).then_return(None)

    del subject[1]


def test_delitem_raises_if_slot_has_module(
    decoy: Decoy,
    mock_protocol_core: ProtocolCore,
    api_version: APIVersion,
    subject: Deck,
) -> None:
    """It should raise a descriptive error if you try to delete a module."""
    decoy.when(mock_protocol_core.robot_type).then_return("OT-3 Standard")
    mock_module_core = decoy.mock(cls=ModuleCore)
    decoy.when(mock_module_core.get_display_name()).then_return("<module display name>")
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(2, api_version, "OT-3 Standard")
    ).then_return(DeckSlotName.SLOT_2)
    decoy.when(mock_protocol_core.get_slot_item(DeckSlotName.SLOT_2)).then_return(
        mock_module_core
    )

    with pytest.raises(
        TypeError,
        match=(
            "Slot 2 contains a module, <module display name>."
            " You can only delete labware, not modules."
        ),
    ):
        del subject[2]


@pytest.mark.parametrize(
    argnames=["slot_definitions_by_name", "staging_slot_definitions_by_name"],
    argvalues=[
        (
            {
                "1": {},
                "2": {},
                "3": {},
            },
            {"4": {}},
        )
    ],
)
def test_slot_keys_iter(subject: Deck) -> None:
    """It should provide an iterable interface to deck slots."""
    result = list(subject)

    assert len(subject) == 3
    assert result == ["1", "2", "3"]


@pytest.mark.parametrize(
    argnames=[
        "slot_definitions_by_name",
        "staging_slot_definitions_by_name",
        "api_version",
    ],
    argvalues=[
        (
            {
                "1": {},
                "2": {},
                "3": {},
            },
            {"4": {}},
            APIVersion(2, 16),
        )
    ],
)
def test_slot_keys_iter_with_staging_slots(subject: Deck) -> None:
    """It should provide an iterable interface to deck slots."""
    result = list(subject)

    assert len(subject) == 4
    assert result == ["1", "2", "3", "4"]


@pytest.mark.parametrize(
    "slot_definitions_by_name",
    [
        {
            "1": {"id": "fee"},
            "2": {"id": "foe"},
            "3": {"id": "fum"},
        }
    ],
)
def test_slots_property(subject: Deck) -> None:
    """It should provide slot definitions."""
    assert subject.slots == [
        {"id": "fee"},
        {"id": "foe"},
        {"id": "fum"},
    ]


@pytest.mark.parametrize(
    "slot_definitions_by_name",
    [
        {
            "2": {
                "id": DeckSlotName.SLOT_2.id,
                "displayName": "foobar",
            }
        }
    ],
)
def test_get_slot_definition(
    decoy: Decoy,
    mock_protocol_core: ProtocolCore,
    api_version: APIVersion,
    subject: Deck,
) -> None:
    """It should provide slot definitions."""
    decoy.when(mock_protocol_core.robot_type).then_return("OT-3 Standard")
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(222, api_version, "OT-3 Standard")
    ).then_return(DeckSlotName.SLOT_2)

    assert subject.get_slot_definition(222) == {
        "id": DeckSlotName.SLOT_2.id,
        "displayName": "foobar",
    }


@pytest.mark.parametrize(
    "slot_definitions_by_name",
    [{"3": {"position": [1.0, 2.0, 3.0]}}],
)
def test_get_position_for(
    decoy: Decoy,
    mock_protocol_core: ProtocolCore,
    api_version: APIVersion,
    subject: Deck,
) -> None:
    """It should return a `Location` for a deck slot."""
    decoy.when(mock_protocol_core.robot_type).then_return("OT-3 Standard")
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(333, api_version, "OT-3 Standard")
    ).then_return(DeckSlotName.SLOT_3)
    decoy.when(
        mock_validation.internal_slot_to_public_string(
            DeckSlotName.SLOT_3, "OT-3 Standard"
        )
    ).then_return("foo")

    result = subject.position_for(333)
    assert result.point == Point(x=1.0, y=2.0, z=3.0)
    assert result.labware.is_slot is True
    assert str(result.labware) == "foo"


def test_highest_z(
    decoy: Decoy, mock_protocol_core: ProtocolCore, subject: Deck
) -> None:
    """It should return the highest Z point of all deck items."""
    decoy.when(mock_protocol_core.get_highest_z()).then_return(42.0)

    assert subject.highest_z == 42.0


def test_right_of_and_left_of(
    decoy: Decoy,
    mock_protocol_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    api_version: APIVersion,
    subject: Deck,
) -> None:
    """It should return items left and right of a slot."""
    left_labware_core = decoy.mock(cls=LabwareCore)
    right_labware_core = decoy.mock(cls=LabwareCore)
    left_labware = decoy.mock(cls=Labware)
    right_labware = decoy.mock(cls=Labware)

    decoy.when(mock_protocol_core.robot_type).then_return("OT-3 Standard")

    decoy.when(mock_adjacent_slots.get_east_slot(4)).then_return(111)
    decoy.when(mock_adjacent_slots.get_west_slot(4)).then_return(999)
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(444, api_version, "OT-3 Standard")
    ).then_return(DeckSlotName.SLOT_4)
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(111, api_version, "OT-3 Standard")
    ).then_return(DeckSlotName.SLOT_1)
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(999, api_version, "OT-3 Standard")
    ).then_return(DeckSlotName.SLOT_9)

    decoy.when(mock_protocol_core.get_slot_item(DeckSlotName.SLOT_1)).then_return(
        right_labware_core
    )
    decoy.when(mock_protocol_core.get_slot_item(DeckSlotName.SLOT_9)).then_return(
        left_labware_core
    )

    decoy.when(mock_core_map.get(right_labware_core)).then_return(right_labware)
    decoy.when(mock_core_map.get(left_labware_core)).then_return(left_labware)

    assert subject.right_of(444) is right_labware
    assert subject.left_of(444) is left_labware


@pytest.mark.parametrize(
    "deck_definition",
    [
        {
            "locations": {
                "orderedSlots": [],
                "calibrationPoints": [
                    {
                        "id": "123",
                        "position": [1.0, 2.0, 3.0],
                        "displayName": "Point 123",
                    },
                    {
                        "id": "456",
                        "position": [4.0, 5.0, 6.0],
                        "displayName": "Point 456",
                    },
                ],
            }
        },
    ],
)
def test_get_calibration_positions(subject: Deck) -> None:
    """It should return calibration positions from the definition."""
    assert subject.calibration_positions == [
        CalibrationPosition(
            id="123", position=(1.0, 2.0, 3.0), displayName="Point 123"
        ),
        CalibrationPosition(
            id="456", position=(4.0, 5.0, 6.0), displayName="Point 456"
        ),
    ]


def test_get_slot_center(
    decoy: Decoy,
    mock_protocol_core: ProtocolCore,
    api_version: APIVersion,
    subject: Deck,
) -> None:
    """It should get the geometric center of a slot."""
    decoy.when(mock_protocol_core.robot_type).then_return("OT-3 Standard")
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(222, api_version, "OT-3 Standard")
    ).then_return(DeckSlotName.SLOT_2)
    decoy.when(mock_protocol_core.get_slot_center(DeckSlotName.SLOT_2)).then_return(
        Point(1, 2, 3)
    )

    assert subject.get_slot_center(222) == Point(1, 2, 3)
